"""Monkey patches C/C++ resolve_xref to save the resolved symbol.

This also other customizations to make use of the resolved symbol.
"""

from typing import Union, Type, Tuple, Optional

import docutils.nodes
import sphinx.addnodes
import sphinx.domains.c
import sphinx.domains.cpp

from sphinx.domains.c import Symbol as CSymbol, CDomain
from sphinx.domains.cpp import Symbol as CPPSymbol, CPPDomain

_last_resolved_symbol: Optional[Union[CSymbol, CPPSymbol]] = None
"""Symbol resolved by `_resolve_xref_inner` or `get_objects`.

This allows additional customizations of those functions to access the symbol.
"""


def get_symbol():
    return _last_resolved_symbol


def set_symbol(symbol: Optional[Union[CSymbol, CPPSymbol]]):
    global _last_resolved_symbol
    _last_resolved_symbol = symbol


def _monkey_patch_resolve_xref_save_symbol(
    symbol_class: Union[Type[CSymbol], Type[CPPSymbol]],
    domain_class: Union[Type[CDomain], Type[CPPDomain]],
):

    orig_resolve_xref_inner = domain_class._resolve_xref_inner

    def _resolve_xref_inner(
        self: Union[CDomain, CPPDomain],
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Tuple[Optional[docutils.nodes.Element], Optional[str]]:
        set_symbol(None)

        orig_find_name = getattr(symbol_class, "find_name", None)

        if orig_find_name:

            def find_name(self: Union[CSymbol, CPPSymbol], *args, **kwargs):
                symbols, failReason = orig_find_name(self, *args, **kwargs)  # type: ignore
                if symbols:
                    set_symbol(symbols[0])
                return symbols, failReason

            symbol_class.find_name = find_name  # type: ignore

        orig_find_declaration = symbol_class.find_declaration

        def find_declaration(self: Union[CSymbol, CPPSymbol], *args, **kwargs):
            symbol = orig_find_declaration(self, *args, **kwargs)  # type: ignore
            if symbol:
                set_symbol(symbol)
            return symbol

        symbol_class.find_declaration = find_declaration  # type: ignore

        try:
            return orig_resolve_xref_inner(
                self, env, fromdocname, builder, typ, target, node, contnode  # type: ignore
            )
        finally:
            if orig_find_name:
                symbol_class.find_name = orig_find_name  # type: ignore
            symbol_class.find_declaration = orig_find_declaration  # type: ignore

    domain_class._resolve_xref_inner = _resolve_xref_inner  # type: ignore


_monkey_patch_resolve_xref_save_symbol(
    sphinx.domains.cpp.Symbol, sphinx.domains.cpp.CPPDomain
)
_monkey_patch_resolve_xref_save_symbol(
    sphinx.domains.c.Symbol, sphinx.domains.c.CDomain
)



"""Sphinx-Immaterial theme."""

import os
from typing import cast, List, Type, Dict, Mapping, Optional, Union

import docutils.nodes
from sphinx.application import Sphinx
import sphinx.builders
import sphinx.builders.html
import sphinx.theming
import sphinx.util.logging
import sphinx.util.fileutil
import sphinx.util.matching
import sphinx.util.docutils
import sphinx.writers.html5

from .apidoc import apidoc_formatting
from . import nav_adapt
from .details_patch import monkey_patch_details_run

logger = sphinx.util.logging.getLogger(__name__)


DEFAULT_THEME_OPTIONS = {
    "features": [],
    "font": {"text": "Roboto", "code": "Roboto Mono"},
    "plugins": {
        "search": {},
    },
    "icon": {},
    "repo_url": "",
    "edit_uri": "",
    "globaltoc_collapse": True,
    "toc_title": None,
    "toc_title_is_page_title": False,
}


def _get_html_translator(
    base_translator: Type[sphinx.writers.html5.HTML5Translator],
) -> Type[sphinx.writers.html5.HTML5Translator]:
    class CustomHTMLTranslator(
        apidoc_formatting.HTMLTranslatorMixin, base_translator  # type: ignore
    ):  # pylint: disable=abstract-method
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            # Ensure pygments uses <code> elements, for compatibility with the
            # mkdocs-material CSS which expects that.
            self.highlighter.formatter_args.update(wrapcode=True)

            # Ensure all tables are marked as data tables.  The material CSS only
            # applies to tables with this class, in order to leave tables used for
            # layout purposes alone.
            self.settings.table_style = ",".join(
                self.settings.table_style.split(",") + ["data"]
            )

            # Ensure classes like `s` (used for string literals in code
            # highlighting) aren't converted to `<s>` elements (strikethrough).
            # Sphinx already overrides this, but for some reason due to
            # `__init__` invocation order it gets overridden.
            self.supported_inline_tags = set()

        def visit_section(
            self, node: Union[docutils.nodes.section, docutils.nodes.Element]
        ) -> None:
            # Sphinx normally writes sections with a section heading as:
            #
            #     <div id="identifier" class="section"><hN>...</hN>...</div>
            #
            # but that is incompatible with the way scroll-margin-top and the
            # `:target` selector are used in the mkdocs-material CSS.  For
            # compatibility with mkdocs-material, we strip the outer `<div>` and
            # instead add the `id` to the inner `<hN>` element.
            #
            # That is accomplished by overriding `visit_section` and
            # `depart_section` not to add the `<div>` and `</div>` tags, and also
            # modifying `visit_title` to insert the `id`.
            self.section_level += 1

        def depart_section(
            self, node: Union[docutils.nodes.section, docutils.nodes.Element]
        ) -> None:
            self.section_level -= 1

        def visit_title(
            self, node: Union[docutils.nodes.title, docutils.nodes.Element]
        ) -> None:
            if isinstance(node.parent, docutils.nodes.section):
                if node.parent.get("ids") and not node.get("ids"):
                    node["ids"] = node.parent.get("ids")
                    super().visit_title(node)
                    del node["ids"]
                    return
            super().visit_title(node)

    return CustomHTMLTranslator


def _get_html_builder(base_builder: Type[sphinx.builders.html.StandaloneHTMLBuilder]):
    """Returns a modified HTML translator."""

    class CustomHTMLBuilder(base_builder):  # type: ignore

        css_files: List[sphinx.builders.html.Stylesheet]
        theme: sphinx.theming.Theme
        templates: sphinx.jinja2glue.BuiltinTemplateLoader

        @property
        def default_translator_class(self):
            return _get_html_translator(super().default_translator_class)

        def init_js_files(self):
            super().init_js_files()

            # Remove unnecessary scripts

            excluded_scripts = set(
                [
                    "_static/underscore.js",
                    "_static/doctools.js",
                    "_static/language_data.js",
                    "_static/documentation_options.js",
                ]
            )
            if nav_adapt.READTHEDOCS is None:
                excluded_scripts.add("_static/jquery.js")
                excluded_scripts.add("_static/_sphinx_javascript_frameworks_compat.js")
            self.script_files = [
                x for x in self.script_files if x.filename not in excluded_scripts
            ]

        def init_css_files(self):
            super().init_css_files()

            # Remove unnecessary styles

            excluded = frozenset(
                [
                    "_static/pygments.css",
                    "_static/material.css",
                    "_static/basic.css",
                ]
            )
            self.css_files = [
                x
                for x in cast(List[sphinx.builders.html.Stylesheet], self.css_files)
                if x.filename not in excluded
            ]

        def gen_additional_pages(self):
            # Prevent the search.html page from being written since this theme provides
            # its own search results display that does not use it.
            search = self.search
            self.search = False
            super().gen_additional_pages()
            self.search = search

        def create_pygments_style_file(self):
            pass

        def copy_theme_static_files(self, context: Dict) -> None:

            # Modified from version in sphinx.builders.html.__init__.py to
            # exclude copying unused static files from `basic` theme.
            def onerror(filename: str, error: Exception) -> None:
                logger.warning(
                    "Failed to copy a file in html_static_file: %s: %r", filename, error
                )

            if self.theme:
                for entry in self.theme.get_theme_dirs()[::-1]:
                    if os.path.basename(entry) == "basic":
                        excluded_list = [
                            "**/.*",
                            "**/doctools.js",
                            "**/underscore*.js",
                            "**/*.png",
                            "**/basic.css_t",
                            "**/documentation_options.js_t",
                            "**/searchtools.js",
                        ]
                        if nav_adapt.READTHEDOCS is None:
                            excluded_list.append("**/jquery*.js")
                            excluded_list.append(
                                "**/_sphinx_javascript_frameworks_compat.js"
                            )
                        excluded = sphinx.util.matching.Matcher(excluded_list)
                    else:
                        excluded = sphinx.util.matching.DOTFILES
                    sphinx.util.fileutil.copy_asset(
                        os.path.join(entry, "static"),
                        os.path.join(self.outdir, "_static"),
                        excluded=excluded,
                        context=context,
                        renderer=cast(
                            sphinx.util.template.BaseRenderer, self.templates
                        ),
                        onerror=onerror,
                    )

        def get_target_uri(self, docname: str, typ: Optional[str] = None) -> str:
            """Strips ``index.html`` suffix from URIs for cleaner links."""
            orig_uri = super().get_target_uri(docname, typ)
            if self.app.config["html_use_directory_uris_for_index_pages"]:
                index_suffix = "index" + self.link_suffix
                if orig_uri == index_suffix:
                    return ""
                if orig_uri.endswith("/" + index_suffix):
                    return orig_uri[: -len(index_suffix)]
            return orig_uri

    return CustomHTMLBuilder


def dict_merge(*dicts: Mapping):
    """Recursively merges the members of one or more dicts."""
    result: dict = {}
    for d in dicts:
        for k, v in d.items():
            if isinstance(v, Mapping) and k in result and isinstance(result[k], dict):
                result[k] = dict_merge(result[k], v)
            else:
                result[k] = v
    return result


def html_page_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: dict,
    doctree: docutils.nodes.Node,
):
    theme_options = app.config["html_theme_options"]
    theme_options = dict_merge(DEFAULT_THEME_OPTIONS, theme_options)

    builder = app.builder
    assert isinstance(builder, sphinx.builders.html.StandaloneHTMLBuilder)

    num_slashes = builder.get_target_uri(pagename).count("/")
    if num_slashes == 0:
        base_url = "."
    else:
        base_url = "/".join(".." for _ in range(num_slashes))

    version_config = None
    if theme_options.get("version_dropdown"):
        version_config = {
            "provider": "mike",
            "staticVersions": theme_options.get("version_info"),
            "versionPath": theme_options.get("version_json"),
        }

    analytics = None
    if theme_options.get("google_analytics"):
        # Parse old-style analytics config for backwards compatibility
        analytics = {
            "provider": "google",  # Google is the only provider currently supported
            "property": theme_options.get("google_analytics")[0],
        }
    if theme_options.get("analytics"):
        analytics = theme_options.get("analytics")

    context.update(
        config=dict_merge(
            context.get("config", {}),
            {
                "theme": theme_options,
                "site_url": theme_options.get("site_url"),
                "site_name": context["docstitle"],
                "repo_url": theme_options.get("repo_url"),
                "repo_name": theme_options.get("repo_name", None),
                "extra": {
                    "version": version_config,
                    "social": theme_options.get("social"),
                    "disqus": theme_options.get("disqus"),
                    "manifest": theme_options.get("pwa_manifest"),
                    "analytics": analytics,
                },
                "plugins": theme_options.get("plugins"),
            },
        ),
        base_url=base_url,
    )


def _builder_inited(app: sphinx.application.Sphinx) -> None:
    # For compatibility with mkdocs
    if isinstance(app.builder, sphinx.builders.html.StandaloneHTMLBuilder):
        # Latex builder does not have a `templates` attribute
        app.builder.templates.environment.filters["url"] = lambda url: url


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:
    """Merge defaults into theme options."""
    if config["language"] is None:
        config["language"] = "en"  # default to English language
    # make code-blocks' line numbers be a separate column of a 1-row table
    config["html_codeblock_linenos_style"] = "table"  # default is "inline"
    config["html_theme_options"] = dict_merge(
        DEFAULT_THEME_OPTIONS, config["html_theme_options"]
    )


def setup(app):
    app.connect("config-inited", _config_inited)

    app.setup_extension(apidoc_formatting.__name__)
    app.setup_extension("sphinx_immaterial.apidoc.python.default")
    app.setup_extension("sphinx_immaterial.apidoc.cpp.default")
    app.setup_extension(nav_adapt.__name__)
    app.setup_extension("sphinx_immaterial.postprocess_html")

    if sphinx.version_info < (5, 0):
        app.setup_extension("sphinx_immaterial.inlinesyntaxhighlight")

    app.setup_extension("sphinx_immaterial.apidoc.object_toc")
    app.setup_extension("sphinx_immaterial.search_adapt")
    app.setup_extension("sphinx_immaterial.apidoc.generic_synopses")

    app.connect("html-page-context", html_page_context)
    app.connect("builder-inited", _builder_inited)
    app.add_config_value(
        "html_use_directory_uris_for_index_pages", False, rebuild="html", types=bool
    )

    app.add_builder(_get_html_builder(app.registry.builders["html"]), override=True)
    app.add_builder(_get_html_builder(app.registry.builders["dirhtml"]), override=True)
    app.add_html_theme("sphinx_immaterial", os.path.abspath(os.path.dirname(__file__)))

    # register our custom adminition directive that are tied to the theme's CSS
    app.setup_extension("sphinx_immaterial.md_admonition")
    app.setup_extension("sphinx_immaterial.content_tabs")
    app.setup_extension("sphinx_immaterial.mermaid_diagrams")
    app.setup_extension("sphinx_immaterial.task_lists")

    # patch the details directive's run method
    monkey_patch_details_run()

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



#!/usr/bin/env python3
"""Applies changes from mkdocs-material to the current directory.

The upstream mkdocs-material revision on which the current directory is based is
assumed to be stored in the MKDOCS_MATERIAL_MERGE_BASE file.

This script computes the differences between that revision and the current
master branch, and then applies them to the current directory via 3-way merge.
If there are conflicts, conflict markers are left in modified files.

Files without conflicts are staged automatically via `git add`.

.. note::
    This script must be run in a Linux shell because:

    1. The use of `git apply` within a context manager (`with` block) results in an
       error saying "Permission Denied". This because git is trying to open a file that
       is already opened by the python interpreter. This is a common obstacle when using
       `tempfile` module on Windows.
    2. There is no `patch` CLI tool native to Windows. Alternatively, we could try using
       a port from the `gnuwin32 <http://gnuwin32.sourceforge.net/packages/patch.htm>`_
       project, but it requires an extra install on user's behalf.
"""

import argparse
import contextlib
import json
import os
import pathlib
import shutil
import subprocess
import tempfile

MKDOCS_EXCLUDE_PATTERNS = [
    # mkdocs-specific configuration files
    ".gitignore",
    ".gitattributes",
    ".github",
    ".browserslistrc",
    ".dockerignore",
    "requirements.txt",
    "setup.py",
    "Dockerfile",
    "MANIFEST.in",
    # Generated files
    "material",
    # mkdocs-specific files
    "src/*.py",
    "src/mkdocs_theme.yml",
    "src/404.html",
    "mkdocs.yml",
    # Unneeded files
    "typings/lunr",
    "src/assets/javascripts/browser/worker",
    "src/assets/javascripts/integrations/search/worker",
    # Files specific to mkdocs' own documentation
    "src/overrides",
    "src/assets/images/favicon.png",
    "src/.icons/logo.*",
    "docs",
    "LICENSE",
    "CHANGELOG",
    "package-lock.json",
    "*.md",
    "giscus.json",
]

ap = argparse.ArgumentParser()
ap.add_argument(
    "--clone-dir",
    default="/tmp/mkdocs-material",
    help="Temporary directory used for pristine checkout of mkdocs-material.  "
    "This remains as a cache after this script completes even if "
    "`--keep-temp` is not specified.",
)
ap.add_argument(
    "--patch-output",
    default="/tmp/mkdocs-material-patch",
    help="Path where patch is written.",
)
ap.add_argument("--source-ref", default="origin/master")
ap.add_argument("--keep-temp", action="store_true", help="Keep temporary workdir")
ap.add_argument(
    "--dry-run", action="store_true", help="Just print the patch but do not apply."
)
args = ap.parse_args()
source_ref = args.source_ref

script_dir = os.path.dirname(__file__)

merge_base_path = os.path.join(script_dir, "MKDOCS_MATERIAL_MERGE_BASE")
merge_base = pathlib.Path(merge_base_path).read_text(encoding="utf-8").strip()

clone_dir = args.clone_dir

if not os.path.exists(clone_dir):
    subprocess.run(
        ["git", "clone", "https://github.com/squidfunk/mkdocs-material", clone_dir],
        check=True,
    )
else:
    subprocess.run(
        ["git", "fetch", "origin"],
        cwd=clone_dir,
        check=True,
    )


def _fix_package_json(path: pathlib.Path) -> None:
    content = json.loads(path.read_text(encoding="utf-8"))
    content.pop("version", None)
    content["dependencies"].pop("lunr")
    content["dependencies"].pop("fuzzaldrin-plus")
    content["dependencies"].pop("lunr-languages")
    content["devDependencies"].pop("@types/lunr")
    content["devDependencies"].pop("@types/fuzzaldrin-plus")
    path.write_text(json.dumps(content, indent=2) + "\n", encoding="utf-8")


def _resolve_ref(ref: str) -> str:
    return subprocess.run(
        ["git", "rev-parse", ref],
        encoding="utf-8",
        cwd=clone_dir,
        check=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()


@contextlib.contextmanager
def _temp_worktree_path():
    if args.keep_temp:
        temp_workdir = tempfile.mkdtemp()
        yield temp_workdir
        return
    with tempfile.TemporaryDirectory() as temp_workdir:
        try:
            yield temp_workdir
        finally:
            subprocess.run(
                ["git", "worktree", "remove", "--force", temp_workdir],
                cwd=clone_dir,
                check=True,
            )


def _create_adjusted_tree(ref: str, temp_workdir: str) -> str:
    print(f"Checking out {source_ref} -> {temp_workdir}")
    subprocess.run(
        ["git", "worktree", "add", "--detach", temp_workdir, ref],
        cwd=clone_dir,
        check=True,
    )
    subprocess.run(
        ["git", "rm", "--quiet", "-r"] + MKDOCS_EXCLUDE_PATTERNS,
        cwd=temp_workdir,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    _fix_package_json(pathlib.Path(temp_workdir) / "package.json")
    try:
        subprocess.run(
            ["git", "commit", "--no-verify", "-a", "-m", "Exclude files"],
            cwd=temp_workdir,
            capture_output=True,
            check=True,
        )
    except subprocess.CalledProcessError as exc:
        # `git commit` fails if user info not in `git config` -> provide verbosity
        raise RuntimeError(str(exc.stderr, encoding="utf-8")) from exc
    return subprocess.run(
        ["git", "rev-parse", "HEAD"],
        cwd=temp_workdir,
        check=True,
        encoding="utf-8",
        stdout=subprocess.PIPE,
    ).stdout.strip()


def _get_git_status(workdir: str):
    status_output = subprocess.run(
        ["git", "status", "--porcelain=v1", "-z", "--no-renames"],
        stdout=subprocess.PIPE,
        check=True,
        text=True,
        cwd=workdir,
    ).stdout
    result = {}
    for line in status_output.split("\x00"):
        if not line:
            continue
        status_code = line[:2]
        filename = line[3:]
        result[filename] = status_code
    return result


def _characterize_git_status(file_status):
    conflicts = []
    updated = []
    for filename, status in file_status.items():
        if "U" in status:
            conflicts.append(filename)
            continue
        if status != "  ":
            updated.append(filename)
    return updated, conflicts


def main():
    resolved_source_ref = _resolve_ref(args.source_ref)
    print(f"SHA for source_ref {args.source_ref} is {resolved_source_ref}")

    print("\nGetting mkdocs-material repo ready")
    with _temp_worktree_path() as temp_workdir:
        new_tree_commit = _create_adjusted_tree(resolved_source_ref, temp_workdir)

    patch_path = os.path.abspath(args.patch_output)
    if not os.path.exists(patch_path):
        os.makedirs(patch_path)
    patch_path += os.sep + "patch_info.diff"

    print("\nGetting sphinx-immaterial repo ready")
    with _temp_worktree_path() as temp_workdir:
        print("    creating a temp workspace")
        old_tree_commit = _create_adjusted_tree(merge_base, temp_workdir)
        subprocess.run(
            ["git", "rm", "-r", "."],
            cwd=temp_workdir,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        print("    copying files to the temp workspace.")
        shutil.copytree(
            script_dir,
            temp_workdir,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns(
                ".git",
                "node_modules",
                ".icons",
                "_build",
            ),
        )
        print("    creating a local-only commit")
        subprocess.run(
            ["git", "add", "-A", "--force", "."],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=temp_workdir,
            check=True,
        )
        subprocess.run(
            ["git", "commit", "--no-verify", "-a", "-m", "Working changes"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=temp_workdir,
            check=True,
        )

        print("\nCreating a diff report")
        with tempfile.NamedTemporaryFile(mode="wb") as patch_f:
            subprocess.run(
                ["git", "diff", f"{old_tree_commit}..{new_tree_commit}"],
                cwd=clone_dir,
                stdout=patch_f,
                check=True,
            )
            patch_f.flush()

            try:
                print("\nCreating a patch report")
                subprocess.run(
                    ["git", "apply", "--3way", patch_f.name],
                    check=True,
                    cwd=temp_workdir,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as exc:
                # provide a verbose coherent output from `git apply` when problematic.
                output = str(exc.stdout, encoding="utf-8").replace("\n", "\n   ")
                output += str(exc.stderr, encoding="utf-8").replace("\n", "\n   ")
                print(f"`{' '.join(exc.cmd)}` returned {exc.returncode}\n   {output}")

        with open(patch_path, "wb") as patch_f:
            subprocess.run(
                ["git", "diff", "HEAD"], check=True, cwd=temp_workdir, stdout=patch_f
            )
        file_status = _get_git_status(temp_workdir)

    updated_files, conflict_files = _characterize_git_status(file_status)

    print("Patch in: " + patch_path)

    if not args.dry_run:
        print("\nApplying patch file to sphinx-immaterial repo.")
        # LINUX ONLY - the `patch` cmd doesn't have a native equivalent on Windows.
        with open(patch_path, "rb") as std_in_file:
            subprocess.run(
                ["patch", "-p1"], stdin=std_in_file, check=True, cwd=script_dir
            )
        print("\nStaging non-conflicting files.")
        subprocess.run(["git", "add", "--"] + updated_files, check=True, cwd=script_dir)
        pathlib.Path(merge_base_path).write_text(
            resolved_source_ref + "\n", encoding="utf-8"
        )
    else:
        print(pathlib.Path(patch_path).read_text(encoding="utf-8"))

    if conflict_files:
        print("File with conflicts:")
        for filename in conflict_files:
            print(f"{file_status[filename]} {filename}")


if __name__ == "__main__":
    main()



"""Fixes ASTDeclarator.isPack method in the Sphinx c++ domain.

https://github.com/sphinx-doc/sphinx/pull/10257
"""

import sphinx.domains.cpp


def _monkey_patch_cpp_is_pack():
    sphinx.domains.cpp.ASTDeclaratorPtr.isPack = property(lambda self: self.next.isPack)
    sphinx.domains.cpp.ASTDeclaratorRef.isPack = property(lambda self: self.next.isPack)
    sphinx.domains.cpp.ASTDeclaratorMemPtr.isPack = property(
        lambda self: self.next.isPack
    )
    sphinx.domains.cpp.ASTDeclaratorParamPack.isPack = property(lambda self: True)
    sphinx.domains.cpp.ASTDeclaratorParen.isPack = property(
        lambda self: self.inner.isPack or self.next.isPack
    )


_monkey_patch_cpp_is_pack()



"""Adaptation of pprint.py from the Python standard library to JSON."""
#  Author:      Fred L. Drake, Jr.
#               fdrake@acm.org

import io
import json


def pformat(obj, indent=1, width=80, depth=None, *, compact=False):
    """Format a Python object into a pretty-printed representation."""
    return _PrettyPrinter(
        indent=indent, width=width, depth=depth, compact=compact
    ).pformat(obj)


class _PrettyPrinter:
    """Implementation of `pformat`."""

    def __init__(self, indent=1, width=80, depth=None, *, compact=False):
        indent = int(indent)
        width = int(width)
        if indent < 0:
            raise ValueError("indent must be >= 0")
        if depth is not None and depth <= 0:
            raise ValueError("depth must be > 0")
        if not width:
            raise ValueError("width must be != 0")
        self._depth = depth
        self._indent_per_level = indent
        self._width = width
        self._compact = bool(compact)

    def pformat(self, obj):
        sio = io.StringIO()
        self._format(obj, sio, 0, 0)
        return sio.getvalue()

    def _repr(self, obj):
        return json.dumps(obj)

    def _format(self, obj, stream, indent, allowance):
        rep = self._repr(obj)
        max_width = self._width - indent - allowance
        if len(rep) > max_width:
            method = None
            if isinstance(obj, dict):
                method = self._pprint_dict
            elif isinstance(obj, list):
                method = self._pprint_list

            if method is not None:
                method(obj, stream, indent, allowance)
                return
        stream.write(rep)

    def _pprint_dict(self, obj, stream, indent, allowance):
        write = stream.write
        write("{")
        length = len(obj)
        if length:
            self._format_dict_items(obj.items(), stream, indent, allowance + 1)
        write("}")

    def _pprint_list(self, obj, stream, indent, allowance):
        stream.write("[")
        self._format_items(obj, stream, indent, allowance + 1)
        stream.write("]")

    def _format_dict_items(self, items, stream, indent, allowance):
        write = stream.write
        indent += self._indent_per_level
        write("\n" + " " * indent)
        delimnl = ",\n" + " " * indent
        last_index = len(items) - 1
        for i, (key, ent) in enumerate(items):
            last = i == last_index
            rep = self._repr(key)
            write(rep)
            write(": ")
            self._format(ent, stream, indent, allowance if last else 1)
            if not last:
                write(delimnl)
            else:
                write("\n" + " " * (indent - self._indent_per_level))

    def _format_items(self, items, stream, indent, allowance):
        write = stream.write
        indent += self._indent_per_level
        if self._indent_per_level > 1:
            write((self._indent_per_level - 1) * " ")
        delimnl = ",\n" + " " * indent
        delim = ""
        width = max_width = self._width - indent + 1
        it = iter(items)
        try:
            next_ent = next(it)
        except StopIteration:
            return
        last = False
        while not last:
            ent = next_ent
            try:
                next_ent = next(it)
            except StopIteration:
                last = True
                max_width -= allowance
                width -= allowance
            if self._compact:
                rep = self._repr(ent)
                w = len(rep) + 2
                if width < w:
                    width = max_width
                    if delim:
                        delim = delimnl
                if width >= w:
                    width -= w
                    write(delim)
                    delim = ", "
                    write(rep)
                    continue
            write(delim)
            delim = delimnl
            self._format(ent, stream, indent, allowance if last else 1)



"""Adds support for `macroParam` as an object type."""

import sphinx.domains.c

from sphinx.domains.c import (
    Symbol as CSymbol,
    ASTMacro,
    ASTMacroParameter,
    ASTDeclaration,
)


def _monkey_patch_c_macro_parameter_symbols():
    """Adds support for the `macroParam` object type to the C domain."""

    orig_add_function_params = CSymbol._add_function_params

    def _add_function_params(self: CSymbol) -> None:
        orig_add_function_params(self)
        if self.declaration is None or not isinstance(
            self.declaration.declaration, ASTMacro
        ):
            return
        args = self.declaration.declaration.args
        if not args:
            return
        for p in args:
            nn = p.arg
            if nn is None:
                continue
            decl = ASTDeclaration("macroParam", None, p)  # type: ignore[arg-type]
            assert not nn.rooted
            assert len(nn.names) == 1
            self._add_symbols(nn, decl, self.docname, self.line)

    CSymbol._add_function_params = _add_function_params

    def get_id(
        self: ASTMacroParameter, version: int, objectType: str, symbol: CSymbol
    ) -> str:
        # the anchor will be our parent
        return symbol.parent.declaration.get_id(version, prefixed=False)

    ASTMacroParameter.get_id = get_id

    sphinx.domains.c.CDomain.object_types["macroParam"] = sphinx.domains.ObjType(
        "macro parameter", "identifier", "var", "member", "data"
    )


_monkey_patch_c_macro_parameter_symbols()



"""Support for cross-referencing externally-defined C++ symbols.

The builtin `sphinx.ext.intersphinx` extension does not work well for C++
because it only supports exact reference matches.  Consequently, a reference to
`std::vector<int>` will not find a more general `std::vector` entry.

This extension implements the simple solution of stripping off all template
arguments when matching, which is not perfect but handles the most common use
cases.
"""

import re
from typing import List, Dict, NamedTuple, Optional

import docutils.nodes
import sphinx.application
import sphinx.environment
import sphinx.addnodes
from typing_extensions import TypedDict


class ObjectInfo(NamedTuple):
    url: str
    desc: str
    object_type: str


class ExternalCppReference(TypedDict):
    url: str
    """URL to use as the target for references to this symbol."""

    object_type: str
    """C++ object type."""

    desc: str
    """Description text to include in the tooltip."""


def _strip_template_arguments(s: str) -> Optional[str]:
    s = s.lstrip(":")
    prev_index = 0
    retained_parts = []
    nested: List[str] = []
    pattern = re.compile("[()<>]")
    s_len = len(s)
    while prev_index < s_len:
        m = pattern.search(s, prev_index)
        if m is None:
            if not nested:
                retained_parts.append(s[prev_index:])
            break
        if not nested:
            retained_parts.append(s[prev_index : m.start()])
        ch = m.group(0)
        if nested and ch == nested[-1]:
            del nested[-1]
            prev_index = m.end()
            continue
        if ch == "(":
            nested.append(")")
            prev_index = m.end()
            continue
        if ch == "<":
            nested.append(">")
            prev_index = m.end()
            continue
        if ch == ")":
            while nested and nested[-1] != ")":
                del nested[-1]
            if not nested:
                # Mismatched parentheses
                return None
            prev_index = m.end()
            continue
        if ch == ">":
            # Unmatched
            prev_index = m.end()
            continue
    return "".join(retained_parts)


def _missing_reference(
    app: sphinx.application.Sphinx,
    env: sphinx.environment.BuildEnvironment,
    node: sphinx.addnodes.pending_xref,
    contnode: docutils.nodes.TextElement,
) -> Optional[docutils.nodes.reference]:

    data = get_mappings(app)

    if node.get("refdomain") != "cpp":
        return None

    target = _strip_template_arguments(node["reftarget"])
    if target is None:
        return None
    result = data.get(target)
    if result is None:
        return None

    reftitle = result.desc

    newnode = docutils.nodes.reference(
        "", "", internal=False, refuri=result.url, reftitle=reftitle
    )
    newnode.append(contnode)
    if result.object_type in ("class", "type alias", "enum"):
        newnode["classes"].append("desctype")

    return newnode


def get_mappings(app: sphinx.application.Sphinx) -> Dict[str, ObjectInfo]:
    env = app.env
    assert env is not None
    cpp_objects = getattr(env, "cppreference_objects", None)
    if cpp_objects is None:
        cpp_objects = {}
        setattr(env, "cppreference_objects", cpp_objects)
        return cpp_objects
    return cpp_objects


def _load_from_config(app: sphinx.application.Sphinx) -> None:
    mappings = get_mappings(app)

    for name, value in app.config.external_cpp_references.items():
        mappings[name] = ObjectInfo(**value)


def setup(app: sphinx.application.Sphinx):
    app.connect("missing-reference", _missing_reference)
    app.connect("builder-inited", _load_from_config)
    app.add_config_value(
        name="external_cpp_references",
        default={},
        types=(Dict[str, "ExternalCppReference"],),
        rebuild="env",
    )
    return {"parallel_read_safe": True, "parallel_write_safe": True}



"""Utilities for use with Sphinx."""

import io
from typing import Optional, Dict, Union, List, Tuple, Mapping, Sequence

import docutils.nodes
import docutils.parsers.rst.states
import docutils.statemachine
import sphinx.addnodes
import sphinx.util.docutils
from typing_extensions import Literal


def to_statemachine_stringlist(
    content: str, source_path: str, source_line: int = 0
) -> docutils.statemachine.StringList:
    """Converts to a docutils StringList with associated source info.

    All lines of `content` are assigned the same source info.

    :param content: Source text.
    :param source_path: Path to the source file, for error messages.
    :param source_line: Line number in source file, for error messages.

    :returns: The string list, which may be passed to `nested_parse`.
    """
    list_lines = docutils.statemachine.string2lines(content)
    items = [(source_path, source_line)] * len(list_lines)
    return docutils.statemachine.StringList(list_lines, items=items)


def format_directive(
    name: str,
    *args: str,
    signatures: Optional[Sequence[str]] = None,
    content: Optional[str] = None,
    options: Optional[Mapping[str, Union[None, str, bool]]] = None,
) -> str:
    """Formats a RST directive into RST syntax.

    :param name: Directive name, e.g. "json:schema".
    :param args: List of directive arguments.
    :param signatures: List of signatures, alternative to ``args``.
    :param content: Directive body content.
    :param options: Directive options.

    :returns: The formatted directive.
    """
    out = io.StringIO()
    out.write("\n\n")
    prefix = f".. {name}:: "
    out.write(prefix)
    assert not args or not signatures
    out.write(" ".join(args))
    if signatures:
        for i, signature in enumerate(signatures):
            if i > 0:
                out.write("\n" + " " * len(prefix))
            out.write(signature)
    out.write("\n")
    if options:
        for key, value in options.items():
            if value is False or value is None:  # pylint: disable=g-bool-id-comparison
                continue
            if value is True:  # pylint: disable=g-bool-id-comparison
                value = ""
            out.write(f"   :{key}: {value}\n")
    if content:
        out.write("\n")
        for line in content.splitlines():
            out.write(f"   {line}\n")
    out.write("\n")
    return out.getvalue()


def append_directive_to_stringlist(
    out: docutils.statemachine.StringList,
    name: str,
    *args: str,
    source_path: str,
    source_line: int,
    indent: int = 0,
    content: Union[str, docutils.statemachine.StringList, None] = None,
    options: Optional[Dict[str, Union[None, str, bool]]] = None,
) -> None:
    """Formats a RST directive into RST syntax and appends to a StringList.

    :param out: StringList to which directive should be appended.
    :param name: Directive name, e.g. "json:schema".
    :param args: List of directive arguments.
    :param source_path: Source path to associate with the first line of the directive
        (and content if specified as a string).
    :param source_line: Source line to associate with the first line of the directive
        (and content if specified as a string).
    :param indent: Base amount to indent every line.
    :param content: Directive body content.
    :param options: Directive options.
    """
    base_indent_str = " " * indent
    # Blank line at beginning to start new directive.
    out.append(base_indent_str, source_path, source_line)
    out.append(
        f"{base_indent_str}.. {name}::" + "".join(f" {arg}" for arg in args),
        source_path,
        source_line,
    )
    content_indent_str = base_indent_str + "   "
    if options:
        for key, value in options.items():
            if value is False or value is None:  # pylint: disable=g-bool-id-comparison
                continue
            if value is True:  # pylint: disable=g-bool-id-comparison
                value = ""
            out.append(f"{content_indent_str}:{key}: {value}", source_path, source_line)
    if isinstance(content, str):
        out.append(base_indent_str, source_path, source_line)
        for line in content.splitlines():
            out.append(content_indent_str + line, source_path, source_line)
    elif content is not None:
        for source, offset, line in content.xitems():
            out.append(content_indent_str + line, source, offset)
    # Blank line at end to denote end of directive.
    out.append(base_indent_str, source_path, source_line)


def parse_rst(
    state: docutils.parsers.rst.states.RSTState,
    text: Union[str, docutils.statemachine.StringList],
    source_path: str = "",
    source_line: int = 0,
) -> List[docutils.nodes.Node]:
    content = (
        to_statemachine_stringlist(text, source_path, source_line)
        if isinstance(text, str)
        else text
    )
    with sphinx.util.docutils.switch_source_input(state, content):
        node = docutils.nodes.container()
        # necessary so that the child nodes get the right source/line set
        node.document = state.document
        state.nested_parse(content, 0, node)
    return node.children


def summarize_element_text(
    node: docutils.nodes.Element,
    mode: Literal["first_paragraph", "first_sentence"] = "first_paragraph",
) -> str:
    """Extracts a short text synopsis, e.g. for use as a tooltip."""

    # Recurisvely extract first paragraph
    while True:
        for p in node.findall(condition=docutils.nodes.paragraph):
            if p is node:
                continue
            node = p
            break
        else:
            break

    text = node.astext()
    if mode == "first_sentence":
        sentence_end = text.find(". ")
        if sentence_end != -1:
            text = text[: sentence_end + 1]
    text = text.replace("\n", " ")
    return text.strip()


def make_toctree_node(
    state: docutils.parsers.rst.states.RSTState,
    toc_entries: List[Tuple[str, str]],
    options: dict,
    source_path: str,
    source_line: int = 0,
) -> List[docutils.nodes.Node]:
    # The Sphinx `toctree` directive parser cannot handle page names that
    # include angle brackets.  Therefore, we use the directive to create an
    # empty toctree node and then add the entries directly.
    toctree_nodes = parse_rst(
        state=state,
        text=format_directive("toctree", options=options),
        source_path=source_path,
        source_line=source_line,
    )
    toctree: Optional[sphinx.addnodes.toctree] = None
    for node in toctree_nodes[-1].findall(condition=sphinx.addnodes.toctree):
        toctree = node
        break
    if toctree is None:
        raise ValueError("No toctree node found")
    toctree["entries"].extend(toc_entries)
    toctree["includefiles"].extend([path for _, path in toc_entries])
    return toctree_nodes



import re

import sphinx.addnodes
import sphinx.application
import sphinx.config


def _maybe_strip_type_annotations(
    app: sphinx.application.Sphinx,
    domain: str,
    objtype: str,
    contentnode: sphinx.addnodes.desc_content,
) -> None:
    if domain != "py":
        return
    obj_desc = contentnode.parent
    assert isinstance(obj_desc, sphinx.addnodes.desc)
    strip_self_type_annotations = app.config.python_strip_self_type_annotations
    strip_return_type_annotations = app.config.python_strip_return_type_annotations
    for signode in obj_desc[:-1]:
        assert isinstance(signode, sphinx.addnodes.desc_signature)
        if strip_self_type_annotations:
            for param in signode.findall(condition=sphinx.addnodes.desc_parameter):
                if param.children[0].astext() == "self":
                    # Remove any annotations on `self`
                    del param.children[1:]
                break
        if strip_return_type_annotations is not None:
            fullname = signode.get("fullname")
            if fullname is None:
                # Python domain failed to parse the signature.  Just ignore it.
                continue
            modname = signode["module"]
            if modname:
                fullname = modname + "." + fullname
            if strip_return_type_annotations.fullmatch(fullname):
                # Remove return type.
                for node in signode.findall(condition=sphinx.addnodes.desc_returns):
                    node.parent.remove(node)


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:

    if (
        config.python_strip_self_type_annotations
        or config.python_strip_return_type_annotations
    ):
        if isinstance(config.python_strip_return_type_annotations, str):
            setattr(
                config,
                "python_strip_return_type_annotations",
                re.compile(config.python_strip_return_type_annotations),
            )
        app.connect("object-description-transform", _maybe_strip_type_annotations)


def setup(app: sphinx.application.Sphinx):
    app.add_config_value(
        "python_strip_self_type_annotations", default=True, rebuild="env", types=(bool,)
    )
    app.add_config_value(
        "python_strip_return_type_annotations",
        default=r".*.(__setitem__|__init__)",
        rebuild="env",
        types=(re.Pattern, type(None)),
    )
    app.connect("config-inited", _config_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Adds support for resolution symbols through type aliases.

https://github.com/sphinx-doc/sphinx/pull/10296
"""

from typing import Optional, Iterator, Any, Union

import sphinx.domains.cpp
from sphinx.domains.cpp import Symbol


def _monkey_patch_cpp_domain_symbol_resolution_through_type_aliases():
    def resolve_type_alias(self: Symbol) -> Optional[Symbol]:
        """Resolves `self` to another symbol if it is a type alias."""
        declaration = self.declaration
        if declaration is None:
            return None
        if declaration.objectType != "type":
            return None
        nested_name: sphinx.domains.cpp.ASTNestedName
        if (
            isinstance(declaration.declaration, sphinx.domains.cpp.ASTTypeUsing)
            and declaration.declaration.type is not None
        ):
            trailing_type_spec = declaration.declaration.type.declSpecs.trailingTypeSpec
            if not isinstance(
                trailing_type_spec, sphinx.domains.cpp.ASTTrailingTypeSpecName
            ):
                return None
            nested_name = trailing_type_spec.name
        elif isinstance(declaration.declaration, sphinx.domains.cpp.ASTType):
            trailing_type_spec = declaration.declaration.declSpecs.trailingTypeSpec
            if not isinstance(
                trailing_type_spec, sphinx.domains.cpp.ASTTrailingTypeSpecName
            ):
                return None
            nested_name = trailing_type_spec.name
        else:
            return None
        symbols, failReason = self.parent.find_name(
            nested_name,
            templateDecls=[],
            typ="any",
            matchSelf=False,
            recurseInAnon=True,
            searchInSiblings=False,
            templateShorthand=True,
        )
        if symbols:
            return symbols[0]
        return None

    def resolve_alias_or_base_type(self) -> Iterator["Symbol"]:
        resolved = resolve_type_alias(self)
        if resolved is not None:
            yield resolved
        declaration = self.declaration
        if declaration is None:
            return
        if declaration.objectType != "class":
            return
        for base in declaration.declaration.bases:
            symbols, failReason = self.parent.find_name(
                base.name,
                templateDecls=[],
                typ="any",
                matchSelf=False,
                recurseInAnon=True,
                searchInSiblings=False,
                templateShorthand=True,
            )
            if symbols:
                yield symbols[0]

    orig_find_named_symbols = Symbol._find_named_symbols

    def _find_named_symbols(
        self: Symbol,
        identOrOp: Union[
            sphinx.domains.cpp.ASTIdentifier, sphinx.domains.cpp.ASTOperator
        ],
        templateParams: Any,
        templateArgs: sphinx.domains.cpp.ASTTemplateArgs,
        templateShorthand: bool,
        matchSelf: bool,
        recurseInAnon: bool,
        correctPrimaryTemplateArgs: bool,
        searchInSiblings: bool,
    ) -> Iterator[Symbol]:
        results = orig_find_named_symbols(
            self,
            identOrOp,
            templateParams,
            templateArgs,
            templateShorthand,
            matchSelf,
            recurseInAnon,
            correctPrimaryTemplateArgs,
            searchInSiblings,
        )
        found_match = False
        for result in results:
            found_match = True
            yield result

        if found_match:
            return

        if not templateShorthand:
            return

        if templateParams is not None or templateArgs is not None:
            # Try match again, ignoring template params and args.
            results = orig_find_named_symbols(
                self,
                identOrOp,
                None,
                None,
                templateShorthand,
                matchSelf,
                recurseInAnon,
                correctPrimaryTemplateArgs,
                searchInSiblings,
            )
            for result in results:
                found_match = True
                yield result

            if found_match:
                return

        # Try to resolve type alias
        for resolved_symbol in resolve_alias_or_base_type(self):
            if resolved_symbol is self:
                continue
            yield from resolved_symbol._find_named_symbols(
                identOrOp,
                templateParams,
                templateArgs,
                templateShorthand,
                matchSelf,
                recurseInAnon,
                correctPrimaryTemplateArgs=False,
                searchInSiblings=False,
            )

    Symbol._find_named_symbols = _find_named_symbols

    in_symbol_lookup_with_shorthand = []

    orig_symbol_lookup = Symbol._symbol_lookup

    def _symbol_lookup(
        self,
        nestedName,
        templateDecls,
        onMissingQualifiedSymbol,
        strictTemplateParamArgLists: bool,
        ancestorLookupType: str,
        templateShorthand: bool,
        matchSelf: bool,
        recurseInAnon: bool,
        correctPrimaryTemplateArgs: bool,
        searchInSiblings: bool,
    ):
        try:
            in_symbol_lookup_with_shorthand.append(templateShorthand)

            return orig_symbol_lookup(
                self,
                nestedName,
                templateDecls,
                onMissingQualifiedSymbol,
                strictTemplateParamArgLists,
                ancestorLookupType,
                templateShorthand,
                matchSelf,
                recurseInAnon,
                correctPrimaryTemplateArgs,
                searchInSiblings,
            )

        finally:
            in_symbol_lookup_with_shorthand.pop()

    Symbol._symbol_lookup = _symbol_lookup

    orig_find_identifier = Symbol.find_identifier

    def find_identifier(
        self,
        identOrOp: Union[
            sphinx.domains.cpp.ASTIdentifier, sphinx.domains.cpp.ASTOperator
        ],
        matchSelf: bool,
        recurseInAnon: bool,
        searchInSiblings: bool,
    ):
        s = orig_find_identifier(
            self, identOrOp, matchSelf, recurseInAnon, searchInSiblings
        )
        if s is not None:
            return s
        if (
            not in_symbol_lookup_with_shorthand
            or not in_symbol_lookup_with_shorthand[-1]
        ):
            return None
        for other in resolve_alias_or_base_type(self):
            if other is self:
                continue
            s = other.find_identifier(
                identOrOp=identOrOp,
                matchSelf=matchSelf,
                recurseInAnon=recurseInAnon,
                searchInSiblings=False,
            )
            if s is not None:
                return s
        return None

    Symbol.find_identifier = find_identifier


_monkey_patch_cpp_domain_symbol_resolution_through_type_aliases()



from typing import List

from sphinx.ext.napoleon.docstring import GoogleDocstring


def _monkey_patch_napoleon_admonition_classes():
    def _add_admonition_class(method_name: str, class_name: str) -> None:
        orig_method = getattr(GoogleDocstring, method_name)

        def wrapper(self: GoogleDocstring, section: str) -> List[str]:
            result = orig_method(self, section)
            result.insert(1, f"   :class: {class_name}")
            return result

        setattr(GoogleDocstring, method_name, wrapper)

    _add_admonition_class("_parse_examples_section", "example")


_monkey_patch_napoleon_admonition_classes()



# Sphinx provides the `code-block` directive for highlighting code blocks.
# Docutils provides the `code` role which in theory can be used similarly by
# defining a custom role for a given programming language:
#
#     .. .. role:: python(code)
#          :language: python
#          :class: highlight
#
# In practice this does not produce correct highlighting because it uses a
# separate highlighting mechanism that results in the "long" pygments class
# names rather than "short" pygments class names produced by the Sphinx
# `code-block` directive and for which this extension contains CSS rules.
#
# In addition, even if that issue is fixed, because the highlighting
# implementation in docutils, despite being based on pygments, differs from that
# used by Sphinx, the output does not exactly match that produced by the Sphinx
# `code-block` directive.
#
# This issue is noted here: //github.com/sphinx-doc/sphinx/issues/5157
#
# This module fixes the problem by modifying the Sphinx HTML translator to
# perform highlighting for the `code` role in the same way as the Sphinx
# `code-block` directive.  The solution is derived from this extension
#
# https://github.com/sphinx-contrib/inlinesyntaxhighlight/blob/master/sphinxcontrib/inlinesyntaxhighlight.py
#
# and from the discussion on this pull request:
#
# https://github.com/sphinx-doc/sphinx/pull/6916

import docutils.parsers.rst.roles
import docutils.parsers.rst
import docutils.nodes
import sphinx.application
import sphinx.writers.html
import sphinx.writers.html5


def code_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """code_role override or create if older docutils used.
    This only creates a literal node without parsing the code. This will
    be done later in sphinx. This override is not really needed, but it
    might give some speed
    """

    docutils.parsers.rst.roles.set_classes(options)

    language = options.get("language", "")
    classes = ["code"]

    if "classes" in options:
        classes.extend(options["classes"])

    if language and language not in classes:
        classes.append(language)

    node = docutils.nodes.literal(rawtext, text, classes=classes, language=language)

    return [node], []


code_role.options = {  # type: ignore
    "class": docutils.parsers.rst.directives.class_option,
    "language": docutils.parsers.rst.directives.unchanged,
}


def _monkey_patch_html_translator(translator_class):
    orig_visit_literal = translator_class.visit_literal

    def visit_literal(self, node: docutils.nodes.literal) -> None:
        lang = node.get("language", None)
        if "code" not in node["classes"] or not lang:
            return orig_visit_literal(self, node)

        def warner(msg):
            self.builder.warn(msg, (self.builder.current_docname, node.line))

        highlight_args = dict(node.get("highlight_args", {}), nowrap=True)
        highlighted = self.highlighter.highlight_block(
            node.astext(), lang, warn=warner, location=node, **highlight_args
        )
        starttag = self.starttag(
            node,
            "code",
            suffix="",
            CLASS="docutils literal highlight highlight-%s" % lang,
        )
        self.body.append(starttag + highlighted.strip() + "</code>")
        raise docutils.nodes.SkipNode

    translator_class.visit_literal = visit_literal
    # Due to the use of `SkipNode`, `depart_literal` is only called if the base
    # (non-highlighting) implementation was used in `visit_literal`.  Therefore,
    # we don't need to override `depart_literal`.


def setup(app: sphinx.application.Sphinx):
    docutils.parsers.rst.roles.register_canonical_role("code", code_role)
    _monkey_patch_html_translator(sphinx.writers.html.HTMLTranslator)
    _monkey_patch_html_translator(sphinx.writers.html5.HTML5Translator)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from typing import Union, Type, Optional, Tuple, Iterator

import docutils.nodes
import sphinx.domains.c
import sphinx.domains.cpp
from sphinx.domains.c import CDomain
from sphinx.domains.cpp import CPPDomain

from . import last_resolved_symbol
from . import parameter_objects
from . import symbol_ids
from .. import apidoc_formatting

SYNOPSIS_ATTR = "_sphinx_immaterial_synopsis"


def set_synopsis(
    symbol: Union[sphinx.domains.cpp.Symbol, sphinx.domains.c.Symbol], synopsis: str
) -> None:
    """Sets the synopsis for a given symbol."""
    setattr(symbol.declaration, SYNOPSIS_ATTR, synopsis)


def _monkey_patch_add_object_type_and_synopsis(
    domain_class: Union[
        Type[sphinx.domains.cpp.CPPDomain], Type[sphinx.domains.c.CDomain]
    ]
):
    """Patch C/C++ resolve_xref to add object type-dependent CSS classes."""
    orig_resolve_xref_inner = domain_class._resolve_xref_inner

    def _resolve_xref_inner(
        self: Union[CDomain, CPPDomain],
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Tuple[Optional[docutils.nodes.Element], Optional[str]]:
        refnode, objtype = orig_resolve_xref_inner(
            self, env, fromdocname, builder, typ, target, node, contnode  # type: ignore
        )
        if refnode is None:
            return refnode, objtype

        assert objtype is not None

        last_symbol = last_resolved_symbol.get_symbol()

        objtype = parameter_objects.get_precise_template_parameter_object_type(
            objtype, last_symbol
        )

        if objtype in (
            "templateTypeParam",
            "templateTemplateParam",
            "class",
            "union",
            "type",
            "enum",
            "concept",
        ):
            refnode["classes"].append("desctype")

        if last_symbol is not None:
            refnode["reftitle"] = apidoc_formatting.format_object_description_tooltip(
                env,
                apidoc_formatting.get_object_description_options(
                    env, self.name, objtype
                ),
                base_title=refnode["reftitle"],
                synopsis=getattr(last_symbol.declaration, SYNOPSIS_ATTR, None),
            )

        return refnode, objtype

    domain_class._resolve_xref_inner = _resolve_xref_inner  # type: ignore


def _monkey_patch_domain_get_object_synopses(
    domain_class: Union[Type[CDomain], Type[CPPDomain]]
):
    def get_object_synopses(
        self: Union[CDomain, CPPDomain],
    ) -> Iterator[Tuple[Tuple[str, str], str]]:
        for symbol in self.data["root_symbol"].get_all_symbols():
            if symbol.declaration is None:
                continue
            assert symbol.docname
            synopsis = getattr(symbol.declaration, SYNOPSIS_ATTR, None)
            if not synopsis:
                continue
            anchor = symbol_ids.get_symbol_anchor(symbol)
            yield ((symbol.docname, anchor), synopsis)

    domain_class.get_object_synopses = get_object_synopses  # type: ignore


_monkey_patch_add_object_type_and_synopsis(sphinx.domains.c.CDomain)
_monkey_patch_add_object_type_and_synopsis(sphinx.domains.cpp.CPPDomain)

_monkey_patch_domain_get_object_synopses(sphinx.domains.c.CDomain)
_monkey_patch_domain_get_object_synopses(sphinx.domains.cpp.CPPDomain)



import sphinx.application

from . import annotation_style
from . import object_ids
from . import synopses
from . import parameter_objects
from . import style_default_values_as_code
from . import subscript_methods
from . import attribute_style
from . import napoleon_admonition_classes
from . import strip_property_prefix
from . import section_titles
from . import autodoc_property_type
from . import type_annotation_transforms
from . import strip_self_and_return_type_annotations


def setup(app: sphinx.application.Sphinx):
    app.setup_extension(parameter_objects.__name__)
    app.setup_extension(strip_property_prefix.__name__)
    app.setup_extension(type_annotation_transforms.__name__)
    app.setup_extension(strip_self_and_return_type_annotations.__name__)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Adds support for #include directives in C/C++ signatures."""

from typing import Type, List

import docutils.nodes
import sphinx.directives
import sphinx.domains.c
import sphinx.domains.cpp


def _monkey_patch_domain_to_support_include_directives(
    object_class: Type[sphinx.directives.ObjectDescription], language: str
):

    orig_get_signatures = object_class.get_signatures

    def is_include(sig: str) -> bool:
        return sig.startswith("#")

    def get_signatures(self: sphinx.directives.ObjectDescription) -> List[str]:
        return [sig for sig in orig_get_signatures(self) if not is_include(sig)]

    object_class.get_signatures = get_signatures  # type: ignore

    orig_run = object_class.run

    def run(self: sphinx.directives.ObjectDescription) -> List[docutils.nodes.Node]:
        nodes = orig_run(self)
        include_directives = [
            sig for sig in orig_get_signatures(self) if is_include(sig)
        ]
        if include_directives:
            obj_desc = nodes[-1]
            assert isinstance(obj_desc, sphinx.addnodes.desc)
            include_sig = sphinx.addnodes.desc_signature("", "")
            include_sig["classes"].append("api-include-path")
            for directive in include_directives:
                container = docutils.nodes.container()
                container += docutils.nodes.literal(
                    directive,
                    directive,
                    classes=[language, "code", "highlight"],
                    language=language,
                )
                include_sig.append(container)

            self.set_source_info(include_sig)
            obj_desc.insert(0, include_sig)
        return nodes

    object_class.run = run  # type: ignore


_monkey_patch_domain_to_support_include_directives(
    object_class=sphinx.domains.cpp.CPPObject, language="cpp"
)
_monkey_patch_domain_to_support_include_directives(
    object_class=sphinx.domains.c.CObject, language="c"
)



import pathlib

import pytest

from sphinx_immaterial.apidoc.python.apigen import _get_api_data

from sphinx.testing.path import path as SphinxPath

pytest_plugins = ("sphinx.testing.fixtures",)


@pytest.fixture
def apigen_make_app(tmp_path: pathlib.Path, make_app):

    conf = """
extensions = [
    "sphinx_immaterial",
    "sphinx_immaterial.apidoc.python.apigen",
]
html_theme = "sphinx_immaterial"
"""

    def make(extra_conf: str = "", **kwargs):
        (tmp_path / "conf.py").write_text(conf + extra_conf, encoding="utf-8")
        (tmp_path / "index.rst").write_text("", encoding="utf-8")
        return make_app(srcdir=SphinxPath(str(tmp_path)), **kwargs)

    yield make


@pytest.mark.parametrize(
    "order_tiebreaker,expected_members",
    [("alphabetical", ["a", "b"]), ("definition_order", ["b", "a"])],
)
def test_alphabetical(apigen_make_app, order_tiebreaker, expected_members):

    app = apigen_make_app(
        confoverrides=dict(
            python_apigen_order_tiebreaker=order_tiebreaker,
            python_apigen_default_groups=[],
            python_apigen_modules={
                "python_apigen_test_modules.alphabetical": "api/",
            },
        ),
    )

    print(app._status.getvalue())
    print(app._warning.getvalue())

    data = _get_api_data(app.env)
    members = [x.name for x in data.top_level_groups["public-members"]]
    assert members == expected_members


def test_classmethod(apigen_make_app):
    testmod = "python_apigen_test_modules.classmethod"
    app = apigen_make_app(
        confoverrides=dict(
            python_apigen_default_groups=[
                ("method:.*", "Methods"),
                ("classmethod:.*", "Class methods"),
                ("staticmethod:.*", "Static methods"),
            ],
            python_apigen_modules={
                testmod: "api/",
            },
        ),
    )

    print(app._status.getvalue())
    print(app._warning.getvalue())

    data = _get_api_data(app.env)

    # FIXME: Currently all methods are assigned to the 'Methods" group because
    # the `classmethod` and `staticmethod` object types aren't actually used by
    # Sphinx.
    #
    # https://github.com/sphinx-doc/sphinx/issues/3743
    assert data.entities[f"{testmod}.Foo.my_method"].group_name == "Methods"
    assert data.entities[f"{testmod}.Foo.my_staticmethod"].group_name == "Methods"
    assert data.entities[f"{testmod}.Foo.my_classmethod"].group_name == "Methods"



"""Fixes limitations of the Sphinx C++ parser."""

from . import fix_cpp_domain_requires_clause  # type: ignore[unused-import]
from . import fix_cpp_domain_symbol_resolution_through_type_aliases  # type: ignore[unused-import]
from . import fix_cpp_is_pack  # type: ignore[unused-import]
from . import fix_cpp_symbol_to_normalize_template_args  # type: ignore[unused-import]



"""Patch C and C++ domain resolve_xref implementation.

This adds support for:

- Resolving C macros from C++ domain xrefs.  This allows a role like
  `cpp:expr` to match both C++ symbols and C macro and macro parameters.
  https://github.com/sphinx-doc/sphinx/issues/10262

- Normal resolution of missing symbols starting with `std::`.  Normally, the
  C++ domain silences warnings about these symbols in a way that blocks
  `missing-reference` handlers from receiving them.
"""


import re
from typing import Optional, Tuple, List, cast

import docutils.nodes
import sphinx.addnodes
import sphinx.domains.c
import sphinx.domains.cpp

from sphinx.domains.cpp import CPPExprRole

# We monkey patch `resolve_xref` multiple times.  We must call
# `_monkey_patch_cpp_resolve_c_xrefs` last, to ensure that the other logic only
# runs once.
from . import last_resolved_symbol  # type: ignore[unused-import]
from . import synopses  # type: ignore[unused-import]

POSSIBLE_MACRO_TARGET_PATTERN = re.compile("^[A-Z]+[A-Z_0-9]*(?:::[a-zA-Z0-9_]+)?$")
"""Pattern for targets that may possibly refer to a macro or macro parameter.

Any targets matching this pattern that are not defined as C++ symbols will be
looked up in the C domain.

For efficiency, this is restricted to names that consist of only uppercase
letters and digits, to avoid a duplicate C domain query for all symbols that
will be resolved by `sphinx_immaterial.apidoc.cpp.external_cpp_references`.

Since macro parameters are named as "<MACRO_NAME>::<PARAMETER_NAME>", this
pattern allows "::" even though it would normally only match C++ symbols.
"""

POSSIBLE_MACRO_PARAMETER_TARGET_PATTERN = re.compile("[^<>:()]+$")
"""Pattern for targets that may refer to a macro parameter within the scope of a
C macro definition.
"""


def _monkey_patch_cpp_resolve_c_xrefs():
    orig_resolve_xref_inner = sphinx.domains.cpp.CPPDomain._resolve_xref_inner

    def _resolve_xref_inner(
        self: sphinx.domains.cpp.CPPDomain,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Tuple[Optional[docutils.nodes.Element], Optional[str]]:

        try:
            refnode, objtype = orig_resolve_xref_inner(
                self, env, fromdocname, builder, typ, target, node, contnode
            )
        except sphinx.errors.NoUri:
            # Only raised for `std::` symbols.  Give `missing-reference` handlers a
            # chance to resolve it.
            return None, None

        if refnode is not None:
            return refnode, objtype

        macro_pattern = getattr(env.app, C_MACRO_PATTERN_ATTR, None)

        if (macro_pattern is not None and macro_pattern.fullmatch(target)) or (
            POSSIBLE_MACRO_PARAMETER_TARGET_PATTERN.fullmatch(target)
            and node.get("c:parent_key")
        ):
            # Give C domain a chance to resolve it.
            c_domain = cast(sphinx.domains.c.CDomain, env.get_domain("c"))
            return c_domain._resolve_xref_inner(
                env, fromdocname, builder, typ, target, node, contnode
            )

        return None, None

    sphinx.domains.cpp.CPPDomain._resolve_xref_inner = _resolve_xref_inner


def _monkey_patch_cpp_expr_role_to_include_c_parent_key():
    """Includes ``c:parent_key`` in pending_xref nodes created by CPPExprRole.

    This allows CPPExprRole to be used from the context of C object descriptions,
    in particular macros.
    """

    orig_run = CPPExprRole.run

    def run(
        self: CPPExprRole,
    ) -> Tuple[List[docutils.nodes.Node], List[docutils.nodes.system_message]]:
        nodes, messages = orig_run(self)
        c_parent_key = self.env.ref_context.get("c:parent_key")
        if c_parent_key is not None:
            for node in nodes:
                for refnode in node.findall(condition=sphinx.addnodes.pending_xref):
                    if refnode.get("refdomain") == "cpp":
                        refnode["c:parent_key"] = c_parent_key
        return nodes, messages

    CPPExprRole.run = run


_monkey_patch_cpp_resolve_c_xrefs()
_monkey_patch_cpp_expr_role_to_include_c_parent_key()


C_MACRO_PATTERN_ATTR = "_sphinx_immaterial_cpp_xref_c_macro_pattern"


def _config_inited(
    app: sphinx.application.Sphinx, config: sphinx.config.Config
) -> None:
    c_macro_pattern = config.cpp_xref_resolve_c_macro_pattern
    # Compute a regular expression that matches a macro name, or macro with
    # `::<param>` suffix.
    setattr(
        app,
        C_MACRO_PATTERN_ATTR,
        re.compile("(?:" + c_macro_pattern + ")(?:::[a-zA-Z0-9_]+)?"),
    )


def setup(app: sphinx.application.Sphinx):

    app.add_config_value(
        "cpp_xref_resolve_c_macro_pattern",
        default="[A-Z]+[A-Z_0-9]*",
        types=(str,),
        rebuild="env",
    )
    app.connect("config-inited", _config_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Normalizes the representation of Symbols for non-specialized templates.

https://github.com/sphinx-doc/sphinx/pull/10257

When declaring e.g.::

    .. cpp:function:: template <typename T> int A<T>::foo()

If there was a previous definition of:

    .. cpp:class:: template <typename T> class A

Then the symbol table will be of the form:

    A: template <typename T> class A
      foo: int foo()

However, if there is no previous definition of the class, then the symbol
table will be of the form (where A has template arguments):

    A: template <typename T> class A<T>
      foo: int foo()

That leads to symbol lookup problems due to the way we remove symbols defined
as part of the "summary".  This monkey patch ensures that the representation
is consistent regardless of whether the class is declared first.
"""

from typing import Union, Any

import sphinx.domains.cpp
from sphinx.domains.cpp import Symbol


def _is_cpp_ast_specialization(
    templateParams: Union[
        sphinx.domains.cpp.ASTTemplateParams, sphinx.domains.cpp.ASTTemplateIntroduction
    ],
    templateArgs: Any,
) -> bool:
    # the names of the template parameters must be given exactly as args
    # and params that are packs must in the args be the name expanded
    if len(templateParams.params) != len(templateArgs.args):
        return True
    # having no template params and no arguments is also a specialization
    if len(templateParams.params) == 0:
        return True
    for i, param in enumerate(templateParams.params):
        arg = templateArgs.args[i]
        # TODO: doing this by string manipulation is probably not the most efficient
        paramName = str(getattr(param, "name"))
        argTxt = str(arg)
        isArgPackExpansion = argTxt.endswith("...")
        if getattr(param, "isPack") != isArgPackExpansion:
            return True
        argName = argTxt[:-3] if isArgPackExpansion else argTxt
        if paramName != argName:
            return True
    return False


def _monkey_patch_cpp_symbol_to_normalize_template_args():
    orig_init = Symbol.__init__

    def __init__(
        self: Symbol,
        parent: Symbol,
        identOrOp: Union[
            sphinx.domains.cpp.ASTIdentifier, sphinx.domains.cpp.ASTOperator
        ],
        templateParams: Union[
            sphinx.domains.cpp.ASTTemplateParams,
            sphinx.domains.cpp.ASTTemplateIntroduction,
        ],
        templateArgs: Any,
        declaration: sphinx.domains.cpp.ASTDeclaration,
        docname: str,
        line: int,
    ) -> None:
        if templateArgs is not None and not _is_cpp_ast_specialization(
            templateParams, templateArgs
        ):
            templateArgs = None
        orig_init(
            self,
            parent,
            identOrOp,
            templateParams,
            templateArgs,
            declaration,
            docname,
            line,
        )

    Symbol.__init__ = __init__


_monkey_patch_cpp_symbol_to_normalize_template_args()



from typing import Optional, List, Tuple, cast, Iterator
import docutils.nodes

import sphinx.addnodes
import sphinx.builders
from sphinx.domains.python import PythonDomain
from sphinx.domains.python import PyObject
import sphinx.environment

from .. import apidoc_formatting
from ... import sphinx_utils


def _monkey_patch_python_domain_to_add_object_synopses_to_references():
    def _add_synopsis(
        self: PythonDomain,
        env: sphinx.environment.BuildEnvironment,
        refnode: docutils.nodes.Element,
    ) -> None:
        name = refnode.get("reftitle")
        obj = self.objects.get(name)
        if obj is None:
            return
        refnode["reftitle"] = apidoc_formatting.format_object_description_tooltip(
            env,
            apidoc_formatting.get_object_description_options(
                env, self.name, obj.objtype
            ),
            base_title=name,
            synopsis=self.data["synopses"].get(name),
        )

    orig_resolve_xref = PythonDomain.resolve_xref

    def resolve_xref(
        self: PythonDomain,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        typ: str,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> Optional[docutils.nodes.Element]:
        refnode = orig_resolve_xref(
            self, env, fromdocname, builder, typ, target, node, contnode
        )
        if refnode is not None:
            _add_synopsis(self, env, refnode)
        return refnode

    PythonDomain.resolve_xref = resolve_xref

    orig_resolve_any_xref = PythonDomain.resolve_any_xref

    def resolve_any_xref(
        self: PythonDomain,
        env: sphinx.environment.BuildEnvironment,
        fromdocname: str,
        builder: sphinx.builders.Builder,
        target: str,
        node: sphinx.addnodes.pending_xref,
        contnode: docutils.nodes.Element,
    ) -> List[Tuple[str, docutils.nodes.Element]]:
        results = orig_resolve_any_xref(
            self, env, fromdocname, builder, target, node, contnode
        )
        for _, refnode in results:
            _add_synopsis(self, env, refnode)
        return results

    PythonDomain.resolve_any_xref = resolve_any_xref


def _monkey_patch_python_domain_to_support_synopses():

    orig_after_content = PyObject.after_content

    orig_transform_content = PyObject.transform_content

    def transform_content(self: PyObject, contentnode) -> None:
        setattr(self, "contentnode", contentnode)
        orig_transform_content(self, contentnode)

    PyObject.transform_content = transform_content

    def after_content(self: PyObject) -> None:
        orig_after_content(self)
        noindex = "noindex" in self.options
        if noindex:
            return

        obj_desc = cast(
            sphinx.addnodes.desc_content, getattr(self, "contentnode")
        ).parent
        signodes = obj_desc.children[:-1]

        py = cast(PythonDomain, self.env.get_domain("py"))

        symbols = []
        for signode in cast(List[docutils.nodes.Element], signodes):
            modname = signode["module"]
            fullname = signode["fullname"]
            symbol = (modname + "." if modname else "") + fullname
            symbols.append(symbol)

        if not symbols:
            return

        options = apidoc_formatting.get_object_description_options(
            self.env, self.domain, self.objtype
        )
        generate_synopses = options["generate_synopses"]
        if generate_synopses is None:
            return
        synopsis = sphinx_utils.summarize_element_text(
            getattr(self, "contentnode"), generate_synopses
        )
        if not synopsis:
            return
        for symbol in symbols:
            py.data["synopses"][symbol] = synopsis

    PyObject.after_content = after_content

    orig_merge_domaindata = PythonDomain.merge_domaindata

    def merge_domaindata(self, docnames: List[str], otherdata: dict) -> None:
        orig_merge_domaindata(self, docnames, otherdata)
        self.data["synopses"].update(otherdata["synopses"])

    PythonDomain.merge_domaindata = merge_domaindata

    def get_object_synopses(
        self: PythonDomain,
    ) -> Iterator[Tuple[Tuple[str, str], str]]:
        synopses = self.data["synopses"]
        for refname, synopsis in synopses.items():
            obj = self.objects.get(refname)
            if not obj:
                continue
            yield ((obj.docname, obj.node_id), synopsis)

    PythonDomain.get_object_synopses = get_object_synopses


_monkey_patch_python_domain_to_add_object_synopses_to_references()
_monkey_patch_python_domain_to_support_synopses()

sphinx.domains.python.PythonDomain.initial_data["synopses"] = {}  # name -> synopsis



"""Modifies the formatting of API documentation."""

import functools
import re
from typing import (
    List,
    TYPE_CHECKING,
    cast,
    Optional,
    Dict,
    Tuple,
    NamedTuple,
    Any,
    Pattern,
    Union,
)

import docutils.nodes
import pydantic
import sphinx.addnodes
import sphinx.application
import sphinx.directives
import sphinx.environment
import sphinx.locale
import sphinx.util.logging
import sphinx.writers.html5
from typing_extensions import Literal

_ = sphinx.locale._

logger = sphinx.util.logging.getLogger(__name__)

ObjectDescriptionOptions = Dict[str, Any]


if TYPE_CHECKING:
    HTMLTranslatorMixinBase = sphinx.writers.html5.HTML5Translator
else:
    HTMLTranslatorMixinBase = object


class HTMLTranslatorMixin(HTMLTranslatorMixinBase):  # pylint: disable=abstract-method
    """Mixin for HTMLTranslator that adds additional CSS classes."""

    def visit_desc(
        self, node: Union[sphinx.addnodes.desc, docutils.nodes.Element]
    ) -> None:
        # Object description node

        # These are converted to `<dl>` elements with the domain and objtype
        # as classes.

        # Augment the list of classes with `objdesc` to make it easier to
        # style these without resorting to hacks.
        node["classes"].append("objdesc")
        super().visit_desc(node)

    def visit_desc_type(
        self, node: Union[sphinx.addnodes.desc_type, docutils.nodes.Element]
    ) -> None:
        self.body.append(
            self.starttag(node, tagname="span", suffix="", CLASS="desctype")
        )

    def depart_desc_type(
        self, node: Union[sphinx.addnodes.desc_type, docutils.nodes.Element]
    ) -> None:
        self.body.append("</span>")

    def visit_desc_parameterlist(
        self, node: Union[sphinx.addnodes.desc_parameterlist, docutils.nodes.Element]
    ) -> None:
        super().visit_desc_parameterlist(node)
        open_paren, _ = node.get("parens", ("(", ")"))
        self.body[-1] = self.body[-1].replace("(", open_paren)

    def depart_desc_parameterlist(
        self, node: Union[sphinx.addnodes.desc_parameterlist, docutils.nodes.Element]
    ) -> None:
        super().depart_desc_parameterlist(node)
        _, close_paren = node.get("parens", ("(", ")"))
        self.body[-1] = self.body[-1].replace(")", close_paren)

    def visit_desc_parameter(
        self, node: Union[sphinx.addnodes.desc_parameter, docutils.nodes.Element]
    ) -> None:
        self.body.append('<span class="sig-param-decl">')
        super().visit_desc_parameter(node)

    def depart_desc_parameter(
        self, node: Union[sphinx.addnodes.desc_parameter, docutils.nodes.Element]
    ) -> None:
        super().depart_desc_parameter(node)
        self.body.append("</span>")

    def depart_field_name(self, node: docutils.nodes.Element) -> None:
        self.add_permalink_ref(node, _("Permalink to this headline"))
        super().depart_field_name(node)

    def depart_rubric(self, node: docutils.nodes.Element) -> None:
        self.add_permalink_ref(node, _("Permalink to this headline"))
        super().depart_rubric(node)

    def depart_term(self, node: docutils.nodes.Element) -> None:
        if "ids" in node.attributes:
            self.add_permalink_ref(node, _("Permalink to this definition"))
        super().depart_term(node)

    def visit_caption(self, node: docutils.nodes.Element) -> None:
        attributes = {"class": "caption-text"}
        if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
            "literal_block"
        ):
            # add highlight class to caption's div container.
            # This is needed to trigger mkdocs-material CSS rule `.highlight .filename`
            self.body.append('<div class="code-block-caption highlight">')
            # append a CSS class to trigger mkdocs-material theme's caption CSS style
            attributes["class"] += " filename"
        else:
            super().visit_caption(node)
        self.add_fignumber(node.parent)
        self.body.append(self.starttag(node, "span", **attributes))

    def depart_caption(self, node: docutils.nodes.Element) -> None:
        if not isinstance(
            node.parent, docutils.nodes.container
        ) and not node.parent.get("literal_block"):
            # only append ending tag if parent is not a literal-block.
            # Because all elements in the caption should be within a span element
            self.body.append("</span>")

        # append permalink if available
        if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
            "literal_block"
        ):
            self.add_permalink_ref(node.parent, _("Permalink to this code"))
            self.body.append("</span>")  # done; add closing tag
        elif isinstance(node.parent, docutils.nodes.figure):
            self.add_permalink_ref(node.parent, _("Permalink to this image"))
        elif node.parent.get("toctree"):
            self.add_permalink_ref(node.parent.parent, _("Permalink to this toctree"))

        if isinstance(node.parent, docutils.nodes.container) and node.parent.get(
            "literal_block"
        ):
            self.body.append("</div>\n")
        else:
            super().depart_caption(node)

    # `desc_inline` nodes are generated by the `cpp:expr` role.
    #
    # Wrap it in a `<code>` element with the "highlight" class to ensure it
    # displays properly as an inline code literal.
    def visit_desc_inline(
        self, node: Union[sphinx.addnodes.desc_inline, docutils.nodes.Element]
    ) -> None:
        self.body.append(
            self.starttag(node, tagname="code", suffix="", CLASS="highlight")
        )

    def depart_desc_inline(
        self, node: Union[sphinx.addnodes.desc_inline, docutils.nodes.Element]
    ) -> None:
        self.body.append("</code>")


def _wrap_signature(node: sphinx.addnodes.desc_signature, limit: int):
    """Wraps long function signatures.

    Adds the `sig-wrap` class which causes each parameter to be displayed on a
    separate line.
    """
    node_text = node.astext()
    if len(node_text) > limit:
        node["classes"].append("sig-wrap")


def _wrap_signatures(
    app: sphinx.application.Sphinx,
    domain: str,
    objtype: str,
    content: docutils.nodes.Element,
) -> None:
    env = app.env
    assert env is not None
    options = get_object_description_options(env, domain, objtype)
    if (
        not options["wrap_signatures_with_css"]
        or options.get("clang_format_style") is not None
    ):
        return
    signatures = content.parent[:-1]
    for signature in signatures:
        assert isinstance(signature, sphinx.addnodes.desc_signature)
        _wrap_signature(signature, options["wrap_signatures_column_limit"])


def _monkey_patch_object_description_to_include_fields_in_toc():
    orig_run = sphinx.directives.ObjectDescription.run

    def run(self: sphinx.directives.ObjectDescription) -> List[docutils.nodes.Node]:
        nodes = orig_run(self)

        options = get_object_description_options(self.env, self.domain, self.objtype)
        if not options["include_fields_in_toc"]:
            return nodes

        obj_desc = nodes[-1]

        obj_id = None
        for sig in obj_desc[:-1]:
            ids = sig["ids"]
            if ids and ids[0]:
                obj_id = ids[0]
                break

        obj_content = obj_desc[-1]
        for child in obj_content:
            if not isinstance(child, docutils.nodes.field_list):
                continue
            for field in child:
                assert isinstance(field, docutils.nodes.field)
                field_name = cast(docutils.nodes.field_name, field[0])
                if field_name["ids"]:
                    continue
                field_id = docutils.nodes.make_id(field_name.astext())
                if obj_id:
                    field_id = f"{obj_id}-{field_id}"
                field_name["ids"].append(field_id)

        return nodes

    sphinx.directives.ObjectDescription.run = run


def _monkey_patch_object_description_to_include_rubrics_in_toc():
    orig_run = sphinx.directives.ObjectDescription.run

    def run(self: sphinx.directives.ObjectDescription) -> List[docutils.nodes.Node]:
        nodes = orig_run(self)

        options = get_object_description_options(self.env, self.domain, self.objtype)
        if not options["include_rubrics_in_toc"]:
            return nodes

        obj_desc = nodes[-1]
        obj_content = obj_desc[-1]
        for child in obj_content:
            if not isinstance(child, docutils.nodes.rubric):
                continue
            rubric = cast(docutils.nodes.rubric, child)
            if rubric["ids"]:
                continue
            rubric_id = docutils.nodes.make_id(rubric.astext())
            rubric["ids"].append(rubric_id)

        return nodes

    sphinx.directives.ObjectDescription.run = run


def format_object_description_tooltip(
    env: sphinx.environment.BuildEnvironment,
    options: ObjectDescriptionOptions,
    base_title: str,
    synopsis: Optional[str],
) -> str:
    title = base_title

    domain = env.get_domain(options["domain"])

    if options["include_object_type_in_xref_tooltip"]:
        object_type = options["object_type"]
        title += f" ({domain.get_type_name(domain.object_types[object_type])})"

    if synopsis:
        title += f" — {synopsis}"

    return title


DEFAULT_OBJECT_DESCRIPTION_OPTIONS: List[Tuple[str, dict]] = [
    ("std:envvar", {"toc_icon_class": "alias", "toc_icon_text": "$"}),
    ("js:module", {"toc_icon_class": "data", "toc_icon_text": "r"}),
    ("js:function", {"toc_icon_class": "procedure", "toc_icon_text": "M"}),
    ("js:method", {"toc_icon_class": "procedure", "toc_icon_text": "M"}),
    ("js:class", {"toc_icon_class": "data", "toc_icon_text": "C"}),
    ("js:data", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("js:attribute", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("json:schema", {"toc_icon_class": "data", "toc_icon_text": "J"}),
    ("json:subschema", {"toc_icon_class": "sub-data", "toc_icon_text": "j"}),
    ("py:class", {"toc_icon_class": "data", "toc_icon_text": "C"}),
    ("py:function", {"toc_icon_class": "procedure", "toc_icon_text": "F"}),
    ("py:method", {"toc_icon_class": "procedure", "toc_icon_text": "M"}),
    ("py:classmethod", {"toc_icon_class": "procedure", "toc_icon_text": "M"}),
    ("py:staticmethod", {"toc_icon_class": "procedure", "toc_icon_text": "M"}),
    ("py:property", {"toc_icon_class": "alias", "toc_icon_text": "P"}),
    ("py:attribute", {"toc_icon_class": "alias", "toc_icon_text": "A"}),
    ("py:data", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("py:parameter", {"toc_icon_class": "sub-data", "toc_icon_text": "p"}),
    ("c:member", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("c:var", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("c:function", {"toc_icon_class": "procedure", "toc_icon_text": "F"}),
    ("c:macro", {"toc_icon_class": "alias", "toc_icon_text": "D"}),
    ("c:union", {"toc_icon_class": "data", "toc_icon_text": "U"}),
    ("c:struct", {"toc_icon_class": "data", "toc_icon_text": "S"}),
    ("c:enum", {"toc_icon_class": "data", "toc_icon_text": "E"}),
    ("c:enumerator", {"toc_icon_class": "data", "toc_icon_text": "e"}),
    ("c:type", {"toc_icon_class": "alias", "toc_icon_text": "T"}),
    (
        "c:macroParam",
        {
            "toc_icon_class": "sub-data",
            "toc_icon_text": "p",
            "generate_synopses": "first_sentence",
        },
    ),
    ("cpp:class", {"toc_icon_class": "data", "toc_icon_text": "C"}),
    ("cpp:struct", {"toc_icon_class": "data", "toc_icon_text": "S"}),
    ("cpp:enum", {"toc_icon_class": "data", "toc_icon_text": "E"}),
    ("cpp:enum-class", {"toc_icon_class": "data", "toc_icon_text": "E"}),
    ("cpp:enum-struct", {"toc_icon_class": "data", "toc_icon_text": "E"}),
    ("cpp:enumerator", {"toc_icon_class": "data", "toc_icon_text": "e"}),
    ("cpp:union", {"toc_icon_class": "data", "toc_icon_text": "U"}),
    ("cpp:concept", {"toc_icon_class": "data", "toc_icon_text": "t"}),
    ("cpp:function", {"toc_icon_class": "procedure", "toc_icon_text": "F"}),
    ("cpp:alias", {"toc_icon_class": "procedure", "toc_icon_text": "F"}),
    ("cpp:member", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("cpp:var", {"toc_icon_class": "alias", "toc_icon_text": "V"}),
    ("cpp:type", {"toc_icon_class": "alias", "toc_icon_text": "T"}),
    ("cpp:namespace", {"toc_icon_class": "alias", "toc_icon_text": "N"}),
    (
        "cpp:functionParam",
        {
            "toc_icon_class": "sub-data",
            "toc_icon_text": "p",
            "generate_synopses": "first_sentence",
        },
    ),
    (
        "cpp:templateTypeParam",
        {
            "toc_icon_class": "alias",
            "toc_icon_text": "T",
            "generate_synopses": "first_sentence",
        },
    ),
    (
        "cpp:templateNonTypeParam",
        {
            "toc_icon_class": "data",
            "toc_icon_text": "N",
            "generate_synopses": "first_sentence",
        },
    ),
    (
        "cpp:templateTemplateParam",
        {
            "toc_icon_class": "alias",
            "toc_icon_text": "T",
            "generate_synopses": "first_sentence",
        },
    ),
    ("rst:directive", {"toc_icon_class": "data", "toc_icon_text": "D"}),
    ("rst:directive:option", {"toc_icon_class": "sub-data", "toc_icon_text": "o"}),
    ("rst:role", {"toc_icon_class": "procedure", "toc_icon_text": "R"}),
]


def get_object_description_option_registry(app: sphinx.application.Sphinx):
    key = "sphinx_immaterial_object_description_option_registry"
    registry = getattr(app, key, None)
    if registry is None:
        registry = {}
        setattr(app, key, registry)
    return registry


class RegisteredObjectDescriptionOption(NamedTuple):
    type_constraint: Any
    default: Any


def add_object_description_option(
    app: sphinx.application.Sphinx, name: str, default: Any, type_constraint: Any = Any
) -> None:
    registry = get_object_description_option_registry(app)
    if name in registry:
        logger.error(f"Object description option {name!r} already registered")
    default = pydantic.parse_obj_as(type_constraint, default)
    registry[name] = RegisteredObjectDescriptionOption(
        default=default, type_constraint=type_constraint
    )


def get_object_description_options(
    env: sphinx.environment.BuildEnvironment, domain: str, object_type: str
) -> ObjectDescriptionOptions:

    return env.app._sphinx_immaterial_get_object_description_options(  # type: ignore
        domain, object_type
    )


def _builder_inited(app: sphinx.application.Sphinx) -> None:

    registry = get_object_description_option_registry(app)
    options_map: Dict[str, List[Tuple[int, Dict[str, Any]]]] = {}
    options_patterns: List[Tuple[Pattern, int, Dict[str, Any]]] = []

    default_options = {}
    for name, registered_option in registry.items():
        default_options[name] = registered_option.default

    # Validate options
    for i, (pattern, options) in enumerate(
        pydantic.parse_obj_as(
            List[Tuple[Pattern, Dict[str, Any]]],
            DEFAULT_OBJECT_DESCRIPTION_OPTIONS + app.config.object_description_options,
        )
    ):
        for name, value in options.items():
            registered_option = registry.get(name)
            if registered_option is None:
                logger.error(
                    "Undefined object description option %r specified for pattern %r",
                    name,
                    pattern.pattern,
                )
                continue
            try:
                options[name] = pydantic.parse_obj_as(
                    registered_option.type_constraint, value
                )
            except Exception as e:  # pylint: disable=broad-except
                logger.error(
                    "Invalid value %r for object description option"
                    " %r specified for pattern %r: %s",
                    value,
                    name,
                    pattern.pattern,
                    e,
                )
        if pattern.pattern == re.escape(pattern.pattern):
            # Pattern just matches a single string.
            options_map.setdefault(pattern.pattern, []).append((i, options))
        else:
            options_patterns.append((pattern, i, options))

    @functools.lru_cache(maxsize=None)
    def get_options(domain: str, object_type: str) -> Dict[str, Any]:
        key = f"{domain}:{object_type}"
        matches = options_map.get(key)
        if matches is None:
            matches = []
        else:
            matches = list(matches)
        for pattern, i, options in options_patterns:
            if pattern.fullmatch(key):
                matches.append((i, options))
        matches.sort(key=lambda x: x[0])
        full_options = default_options.copy()
        for _, m in matches:
            full_options.update(m)
        full_options.update(domain=domain, object_type=object_type)
        return full_options

    app._sphinx_immaterial_get_object_description_options = get_options  # type: ignore


def setup(app: sphinx.application.Sphinx):
    """Registers the monkey patches.

    Does not register HTMLTranslatorMixin, the caller must do that.
    """
    # Add "highlight" class in order for pygments syntax highlighting CSS rules
    # to apply.
    sphinx.addnodes.desc_signature.classes.append("highlight")

    app.connect("object-description-transform", _wrap_signatures, priority=1000)
    _monkey_patch_object_description_to_include_fields_in_toc()
    _monkey_patch_object_description_to_include_rubrics_in_toc()

    add_object_description_option(
        app, "wrap_signatures_with_css", type_constraint=bool, default=True
    )
    add_object_description_option(
        app, "wrap_signatures_column_limit", type_constraint=int, default=68
    )
    add_object_description_option(
        app, "include_in_toc", type_constraint=bool, default=True
    )
    add_object_description_option(
        app, "include_fields_in_toc", type_constraint=bool, default=True
    )
    add_object_description_option(
        app, "include_rubrics_in_toc", type_constraint=bool, default=False
    )
    add_object_description_option(
        app,
        "generate_synopses",
        type_constraint=Optional[Literal["first_paragraph", "first_sentence"]],
        default="first_paragraph",
    )
    add_object_description_option(
        app, "include_object_type_in_xref_tooltip", type_constraint=bool, default=True
    )
    add_object_description_option(
        app, "toc_icon_text", type_constraint=Optional[str], default=None
    )
    add_object_description_option(
        app, "toc_icon_class", type_constraint=Optional[str], default=None
    )

    app.add_config_value(
        "object_description_options",
        default=[],
        rebuild="env",
        types=(List[Tuple[Pattern, Dict[str, Any]]],),
    )
    app.connect("builder-inited", _builder_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import multiprocessing
from typing import cast, Any

from xml.etree import ElementTree

import docutils.nodes
import sphinx.application
import sphinx.builders.html
import sphinx.util.console


def add_html_link(
    app: sphinx.application.Sphinx,
    pagename: str,
    templatename: str,
    context: dict,
    doctree: docutils.nodes.Node,
):
    """As each page is built, collect page names for the sitemap"""
    base_url = app.config["html_theme_options"].get("site_url", "")
    builder = app.builder
    assert isinstance(builder, sphinx.builders.html.StandaloneHTMLBuilder)
    if base_url:
        if not base_url.endswith("/"):
            base_url += "/"
        full_url = base_url + builder.get_target_uri(pagename)
        cast(Any, app).sitemap_links.append(full_url)


def create_sitemap(app: sphinx.application.Sphinx, exception):
    """Generates the sitemap.xml from the collected HTML page links"""
    sitemap_links = cast(Any, app).sitemap_links

    if (
        not app.config["html_theme_options"].get("site_url", "")
        or exception is not None
        or not sitemap_links
    ):
        return

    filename = app.outdir + "/sitemap.xml"
    print(
        "Generating sitemap for {0} pages in "
        "{1}".format(len(sitemap_links), sphinx.util.console.colorize("blue", filename))
    )

    root = ElementTree.Element("urlset")
    root.set("xmlns", "http://www.sitemaps.org/schemas/sitemap/0.9")

    for link in sitemap_links:
        url = ElementTree.SubElement(root, "url")
        ElementTree.SubElement(url, "loc").text = link
    sitemap_links[:] = []

    ElementTree.ElementTree(root).write(filename)


def setup(app: sphinx.application.Sphinx):
    app.connect("html-page-context", add_html_link)
    app.connect("build-finished", create_sitemap)
    manager = multiprocessing.Manager()
    cast(Any, app).sitemap_links = manager.list()
    setattr(app, "multiprocess_manager", manager)
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Optional type annotation transformations."""

import functools
import re
import sys
from typing import (
    cast,
    Dict,
    Union,
    Optional,
    Tuple,
    Sequence,
    NamedTuple,
    Pattern,
    List,
)

import docutils.nodes
import sphinx.application
import sphinx.domains.python
import sphinx.environment
from sphinx.pycode.ast import ast
import sphinx.util.logging

logger = sphinx.util.logging.getLogger(__name__)

PEP585_ALIASES = {
    "typing.Tuple": "tuple",
    "typing.List": "list",
    "typing.Dict": "dict",
    "typing.DefaultDict": "collections.defaultdict",
    "typing.OrderedDict": "collections.OrderedDict",
    "typing.Counter": "collections.Counter",
    "typing.Set": "set",
    "typing.Callable": "collections.abc.Callable",
    "typing.Deque": "collections.deque",
    "typing.FrozenSet": "frozenset",
    "typing.Type": "type",
    "typing.AbstractSet": "collections.abc.Set",
    "typing.ByteString": "collections.abc.ByteString",
    "typing.Collection": "collections.abc.Collection",
    "typing.Container": "collections.abc.Container",
    "typing.ItemsView": "collections.abc.ItemsView",
    "typing.KeysView": "collections.abc.KeysView",
    "typing.Mapping": "collections.abc.Mapping",
    "typing.MappingView": "collections.abc.MappingView",
    "typing.MutableMapping": "collections.abc.MutableMapping",
    "typing.MutableSequence": "collections.abc.MutableSequence",
    "typing.MutableSet": "collections.abc.MutableSet",
    "typing.Sequence": "collections.abc.Sequence",
    "typing.ValuesView": "collections.abc.ValuesView",
    "typing.Iterable": "collections.abc.Iterable",
    "typing.Iterator": "collections.abc.Iterator",
    "typing.Generator": "collections.abc.Generator",
    "typing.Hashable": "collections.abc.Hashable",
    "typing.Reversible": "collections.abc.Reversible",
    "typing.Sized": "collections.abc.Sized",
    "typing.Coroutine": "collections.abc.Coroutine",
    "typing.AsyncGenerator": "collections.abc.AsyncGenerator",
    "typing.AsyncIterable": "collections.abc.AsyncIterable",
    "typing.AsyncIterator": "collections.abc.AsyncIterator",
    "typing.Awaitable": "collections.abc.Awaitable",
    "typing.ContextManager": "contextlib.AbstractContextManager",
    "typing.AsyncContextManager": "contextlib.AbstractAsyncContextManager",
    "typing.Pattern": "re.Pattern",
    "typing.re.Pattern": "re.Pattern",
    "typing.Match": "re.Match",
    "typing.re.Match": "re.Match",
}

TYPING_NAMES = [
    "Any",
    "Tuple",
    "Optional",
    "Union",
    "Literal",
    "Callable",
    "List",
    "Dict",
    "Set",
    "FrozenSet",
    "Sequence",
    "Type",
]


def _get_ast_dotted_name(node: Union[ast.Name, ast.Attribute]) -> Optional[str]:
    if isinstance(node, ast.Name):
        return node.id
    parent = _get_ast_dotted_name(cast(Union[ast.Name, ast.Attribute], node.value))
    if parent is None:
        return None
    return f"{parent}.{node.attr}"


def _dotted_name_to_ast(
    name: str, ctx: ast.expr_context
) -> Union[ast.Name, ast.Attribute]:
    parts = name.split(".")
    tree: Union[ast.Name, ast.Attribute] = ast.Name(parts[0], ctx)
    for part in parts[1:]:
        tree = ast.Attribute(tree, part, ctx)
    return tree


if sys.version_info < (3, 8):
    _CONSTANT_AST_NODE_TYPES = (
        ast.Constant,
        ast.Num,
        ast.Str,
        ast.Bytes,
        ast.Ellipsis,
        ast.NameConstant,
    )
else:
    _CONSTANT_AST_NODE_TYPES = (ast.Constant,)


_CONFIG_ATTR = "_sphinx_immaterial_python_type_transform_config"


class TypeTransformConfig(NamedTuple):
    transform: bool
    transform_names: bool
    aliases: Dict[str, str]
    module_aliases: Dict[str, str]
    module_replacements_pattern: Optional[Pattern[str]]
    concise_literal: bool
    pep604: bool
    strip_modules_from_xrefs_pattern: Optional[Pattern[str]]

    def transform_dotted_name(self, dotted_name: str) -> str:
        aliases = self.aliases
        if aliases:
            new_name = aliases.get(dotted_name)
            if new_name is not None:
                return new_name
        module_replacements_pattern = self.module_replacements_pattern
        if module_replacements_pattern:
            dotted_name = module_replacements_pattern.sub(
                lambda m: self.module_aliases[m.group(0)], dotted_name
            )
        return dotted_name


def _retain_explicit_literal(node: ast.AST) -> bool:
    """Checks if the concise literal syntax cannot be used.

    Since constants cannot be types, there is no ambiguity.
    """
    return not isinstance(node, _CONSTANT_AST_NODE_TYPES)


class TypeAnnotationTransformer(ast.NodeTransformer):
    """Transforms the AST of a type annotation to improve readability.

    - Converts Union/Optional/Literal to "|" syntax allowed by PEP 604.  This
      syntax is not actually supported until Python 3.10 but displays better in
      the documentation.
    """

    config: TypeTransformConfig

    def visit_Name(self, node: ast.Name) -> ast.AST:
        if self.config.transform_names:
            dotted_name = node.id
            new_dotted_name = self.config.transform_dotted_name(dotted_name)
            if dotted_name != new_dotted_name:
                return _dotted_name_to_ast(new_dotted_name, node.ctx)
        return self.generic_visit(node)

    def visit_Attribute(self, node: ast.Attribute) -> ast.AST:
        if self.config.transform_names:
            dotted_name = _get_ast_dotted_name(node)
            if dotted_name is not None:
                new_dotted_name = self.config.transform_dotted_name(dotted_name)
                if new_dotted_name != dotted_name:
                    return _dotted_name_to_ast(new_dotted_name, node.ctx)
        return self.generic_visit(node)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> ast.AST:
        if not isinstance(node.op, ast.Invert) or not isinstance(
            node.operand, ast.Subscript
        ):
            return self.generic_visit(node)
        operand, pep604_transformed = self._transform_subscript_pep604(node.operand)
        if pep604_transformed:
            return operand
        return ast.UnaryOp(node.op, operand)

    def _transform_subscript_pep604(self, node: ast.Subscript) -> Tuple[ast.AST, bool]:
        if not self.config.pep604:
            return self.generic_visit(node), False
        value = self.visit(node.value)
        id_str = _get_ast_dotted_name(value)
        if id_str in ("typing.Optional", "typing.Union", "typing.Literal"):
            slice_expr: ast.AST = node.slice
            elts: Sequence[ast.AST]
            if isinstance(slice_expr, ast.Index):
                slice_expr = slice_expr.value  # type: ignore
            if isinstance(slice_expr, ast.Tuple):
                elts = slice_expr.elts
            else:
                elts = [slice_expr]
            elts = [self.visit(x) for x in elts]
            if id_str == "typing.Optional":
                elts.append(ast.Constant(value=None))
            elif id_str == "typing.Literal":
                elts = [
                    ast.Subscript(value, x, node.ctx)
                    if not self.config.concise_literal or _retain_explicit_literal(x)
                    else x
                    for x in elts
                ]
            result = functools.reduce(
                (lambda left, right: ast.BinOp(left, ast.BitOr(), right)), elts
            )
            return result, True
        return ast.Subscript(value, self.visit(node.slice), node.ctx), False

    def visit_Subscript(self, node: ast.Subscript) -> ast.AST:
        return self._transform_subscript_pep604(node)[0]


def _monkey_patch_python_domain_to_transform_type_annotations():

    orig_parse_annotation = sphinx.domains.python._parse_annotation

    def _parse_annotation(annotation: str, env: sphinx.environment.BuildEnvironment):
        transformer_config = getattr(env, _CONFIG_ATTR, None)
        if transformer_config is None or not transformer_config.transform:
            return orig_parse_annotation(annotation, env)
        try:
            orig_ast_parse = sphinx.domains.python.ast_parse

            def ast_parse(annotation: str) -> ast.AST:
                tree = orig_ast_parse(annotation)
                transformer = TypeAnnotationTransformer()
                transformer.config = cast(TypeTransformConfig, transformer_config)
                return ast.fix_missing_locations(transformer.visit(tree))

            sphinx.domains.python.ast_parse = ast_parse  # type: ignore
            return orig_parse_annotation(annotation, env)
        finally:
            sphinx.domains.python.ast_parse = orig_ast_parse

    sphinx.domains.python._parse_annotation = _parse_annotation


def _builder_inited(app: sphinx.application.Sphinx):
    config = app.config

    aliases: Dict[str, str] = {}
    module_aliases: Dict[str, str] = {}

    if config.python_transform_type_annotations_pep585:
        aliases.update(PEP585_ALIASES)

    if config.python_resolve_unqualified_typing:
        for name in TYPING_NAMES:
            full_name = f"typing.{name}"
            aliases[name] = aliases.get(full_name, full_name)

    if config.python_transform_typing_extensions:
        module_aliases["typing_extensions."] = "typing."

    for source_name, target_name in config.python_type_aliases.items():
        if source_name.endswith("."):
            if target_name and not target_name.endswith("."):
                logger.error(
                    "Invalid python_type_aliases entry %r -> %r: "
                    "source names ending in '.' must map to either "
                    "the empty string or a target name ending in '.'",
                    source_name,
                    target_name,
                )
                continue
            module_aliases[source_name] = target_name
        else:
            aliases[source_name] = target_name

    module_replacements_pattern: Optional[Pattern[str]] = None

    if module_aliases:
        module_replacements_pattern = re.compile(
            "|".join(f"(?:^{re.escape(name)})" for name in module_aliases)
        )

    if (
        config.python_transform_type_annotations_concise_literal
        and not config.python_transform_type_annotations_pep604
    ):
        logger.error(
            "python_transform_type_annotations_concise_literal=True "
            "requires python_transform_type_annotations_pep604=True"
        )

    strip_modules_from_xrefs_pattern: Optional[Pattern[str]] = None

    if config.python_module_names_to_strip_from_xrefs:
        strip_modules_from_xrefs_pattern = re.compile(
            "|".join(
                f"(?:^{re.escape(x)}\\.)"
                for x in config.python_module_names_to_strip_from_xrefs
            )
        )

    setattr(
        app.env,
        _CONFIG_ATTR,
        TypeTransformConfig(
            transform=bool(
                aliases
                or config.python_transform_type_annotations_pep604
                or module_aliases
            ),
            transform_names=bool(aliases or module_aliases),
            aliases=aliases,
            module_aliases=module_aliases,
            module_replacements_pattern=module_replacements_pattern,
            concise_literal=config.python_transform_type_annotations_concise_literal,
            pep604=config.python_transform_type_annotations_pep604,
            strip_modules_from_xrefs_pattern=strip_modules_from_xrefs_pattern,
        ),
    )


def _monkey_patch_python_domain_to_transform_xref_titles():
    orig_type_to_xref = sphinx.domains.python.type_to_xref

    def type_to_xref(
        target: str,
        env: sphinx.environment.BuildEnvironment,
        suppress_prefix: bool = False,
    ) -> sphinx.addnodes.pending_xref:
        node = orig_type_to_xref(target, env, suppress_prefix)
        if (
            not suppress_prefix
            and len(node.children) == 1
            and node.children[0].astext() == target
        ):
            transformer_config = getattr(env, _CONFIG_ATTR, None)
            if transformer_config is not None:
                strip_modules_from_xrefs_pattern = (
                    transformer_config.strip_modules_from_xrefs_pattern
                )
                if strip_modules_from_xrefs_pattern is not None:
                    new_target = strip_modules_from_xrefs_pattern.sub("", target)
                    if new_target != target:
                        del node.children[:]
                        node.append(docutils.nodes.Text(new_target))
        return node

    sphinx.domains.python.type_to_xref = type_to_xref


def setup(app: sphinx.application.Sphinx):
    _monkey_patch_python_domain_to_transform_type_annotations()
    _monkey_patch_python_domain_to_transform_xref_titles()

    app.add_config_value(
        "python_type_aliases",
        default={},
        types=(Dict[str, str],),
        rebuild="env",
    )

    app.add_config_value(
        "python_resolve_unqualified_typing",
        default=True,
        types=(bool,),
        rebuild="env",
    )

    app.add_config_value(
        "python_transform_typing_extensions",
        default=True,
        types=(bool,),
        rebuild="env",
    )

    app.add_config_value(
        "python_transform_type_annotations_pep585",
        default=True,
        types=(bool,),
        rebuild="env",
    )

    app.add_config_value(
        "python_transform_type_annotations_pep604",
        default=True,
        types=(bool,),
        rebuild="env",
    )

    app.add_config_value(
        "python_transform_type_annotations_concise_literal",
        default=True,
        types=(bool,),
        rebuild="env",
    )

    app.add_config_value(
        "python_module_names_to_strip_from_xrefs",
        default=[],
        types=(List[str],),
        rebuild="env",
    )

    app.connect("builder-inited", _builder_inited)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""A custom directive that allows using mermaid diagrams"""
from typing import List
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective
from sphinx.application import Sphinx


class mermaid_node(nodes.General, nodes.Element):
    pass


class MermaidDirective(SphinxDirective):
    """a special directive"""

    has_content = True
    option_spec = {
        "name": directives.unchanged,
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        content = "\n".join(self.content)
        diagram = mermaid_node("", classes=["mermaid"], content=content)
        diagram += nodes.literal("", content, format="html")
        diagram_div = nodes.container(
            "",
            is_div=True,
            classes=["mermaid-diagram"] + self.options.get("class", []),
        )
        if self.options.get("name", ""):
            self.add_name(diagram_div)
        diagram_div += diagram
        self.set_source_info(diagram_div)
        return [diagram_div]


def visit_mermaid_node_html(self, node: mermaid_node):
    attributes = {"class": "mermaid"}
    self.body.append(self.starttag(node, "pre", **attributes))


def depart_mermaid_node_html(self, node: mermaid_node):
    self.body.append("</pre>")


def visit_mermaid_node_latex(self, node: mermaid_node):
    self.body.append("\n\\begin{sphinxVerbatim}[commandchars=\\\\\\{\\}]\n")


def depart_mermaid_node_latex(self, node: mermaid_node):
    self.body.append("\n\\end{sphinxVerbatim}\n")


def setup(app: Sphinx):
    app.add_directive("md-mermaid", MermaidDirective)
    app.add_node(
        mermaid_node,
        html=(visit_mermaid_node_html, depart_mermaid_node_html),
        latex=(visit_mermaid_node_latex, depart_mermaid_node_latex),
    )
    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }
